# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
from collections import namedtuple

import numpy as np

from .. import Device, Computer
from ...constants import EventConstants, DeviceConstants
from ...constants import MouseConstants, KeyboardConstants
from ...errors import print2err, printExceptionDetailsToStdErr
from psychopy.tools.monitorunittools import cm2pix, deg2pix, pix2cm, pix2deg

RectangleBorder = namedtuple('RectangleBorderClass', 'left top right bottom')
currentSec = Computer.getTime

# OS 'independent' view of the Mouse Device

class MouseDevice(Device):
    """
    """
    EVENT_CLASS_NAMES = [
        'MouseInputEvent',
        'MouseButtonEvent',
        'MouseScrollEvent',
        'MouseMoveEvent',
        'MouseDragEvent',
        'MouseButtonPressEvent',
        'MouseButtonReleaseEvent',
        'MouseMultiClickEvent']

    DEVICE_TYPE_ID = DeviceConstants.MOUSE
    DEVICE_TYPE_STRING = 'MOUSE'

    __slots__ = [
        '_lock_mouse_to_display_id',
        '_scrollPositionY',
        '_position',
        '_clipRectsForDisplayID',
        '_lastPosition',
        '_display_index',
        '_last_display_index',
        '_isVisible',
        'activeButtons']

    def __init__(self, *args, **kwargs):
        Device.__init__(self, *args, **kwargs)
        self._clipRectsForDisplayID = {}
        self._lock_mouse_to_display_id = None
        self._scrollPositionY = 0
        self._position = None
        self._lastPosition = None
        self._isVisible = 0
        self._display_index = None
        self._last_display_index = None
        self.activeButtons = {
            MouseConstants.MOUSE_BUTTON_LEFT: 0,
            MouseConstants.MOUSE_BUTTON_RIGHT: 0,
            MouseConstants.MOUSE_BUTTON_MIDDLE: 0,
        }

    def getCurrentButtonStates(self):
        """Returns a list of three booleans, representing the current state of
        the MOUSE_BUTTON_LEFT, MOUSE_BUTTON_MIDDLE, MOUSE_BUTTON_RIGHT ioHub
        Mouse Device.

        Args:
            None

        Returns:
            (left_pressed, middle_pressed, right_pressed): Tuple of 3 bool
            values where True == Pressed.

        """
        return (self.activeButtons[MouseConstants.MOUSE_BUTTON_LEFT] != 0,
                self.activeButtons[MouseConstants.MOUSE_BUTTON_MIDDLE] != 0,
                self.activeButtons[MouseConstants.MOUSE_BUTTON_RIGHT] != 0)

    def _initialMousePos(self):
        """If getPosition is called prior to any mouse events being received,
        this method gets the current system cursor pos using ctypes.
        """
        if self._position is None:
            print2err('ERROR: _initialMousePos must be overwritten by '
                      'OS dependent implementation')

    def getPosition(self, return_display_index=False):
        """Returns the current position of the ioHub Mouse Device. Mouse
        Position is in display coordinate units, with 0,0 being the center of
        the screen.

        Args:
            return_display_index: If True, the display index that is
            associated with the mouse position will also be returned.

        Returns:
            tuple: If return_display_index is false (default), return (x,
            y) position of mouse. If return_display_index is True return ( (
            x,y), display_index).

        """
        self._initialMousePos()
        if return_display_index is True:
            return (tuple(self._position), self._display_index)
        return tuple(self._position)


    def setPosition(self, pos, display_index=None):
        """Sets the current position of the ioHub Mouse Device. Mouse position
        ( pos ) should be specified in Display coordinate units, with 0,0 being
        the center of the screen.

        Args:
            pos ( (x,y) list or tuple ): The position, in Display
            coordinate space, to set the mouse position too.

            display_index (int): Optional argument giving the display index
            to set the mouse pos within. If None, the active ioHub Display
            device index is used.

        Returns:
            tuple: new (x,y) position of mouse in Display coordinate space.
        """
        try:
            pos = pos[0], pos[1]
        except Exception:
            print2err('Warning: Mouse.setPosition: pos must be a list of '
                      'two numbers, not: ', pos)
            return self._position

        display = self._display_device
        if display_index is None:
            display_index = display.getIndex()

        if 0 > display_index >= display.getDisplayCount():
            print2err('Warning: Mouse.setPosition({},{}) failed. '
                      'Display Index must be between '
                      '0 and {}.'.format(pos, display_index,
                                         display.getDisplayCount()-1))
            return self._position

        px, py = display._displayCoord2Pixel(pos[0], pos[1], display_index)

        if not self._isPixPosWithinDisplay((px, py), display_index):
            print2err('Warning: Mouse.setPosition({},{}) failed because '
                      'requested position ({} pix) does not fall within '
                      'specified display pixel bounds.'.format(pos,
                                                               display_index,
                                                               (px, py)))
            return self._position


        mouse_display_index = self.getDisplayIndexForMousePosition((px, py))
        if mouse_display_index != display_index:
            print2err(
                    ' !!! requested display_index {0} != mouse_display_index '
                    '{1}'.format(
                            display_index, mouse_display_index))
            print2err(' mouse.setPos did not update mouse pos')
        else:
            self._lastPosition = self._position
            self._position = pos[0], pos[1]

            self._last_display_index = self._display_index
            self._display_index = mouse_display_index

            self._nativeSetMousePos(px, py)
        return self._position

    def _desktopToWindowPos(self, dpos):
        winfos = self._iohub_server._psychopy_windows
        for w in winfos.values():
            mx, my = dpos
            wx, wy = w['pos'][0],  w['pos'][1]
            ww, wh = w['size'][0], w['size'][1]
            if w['useRetina']:
                ww = ww / 2
                wh = wh / 2
            if wx <= mx <= wx+ww:
                if wy <= my <= wy + wh:
                    return w['handle'], mx - wx - ww/2, -(my - wy - wh/2)
        return None, None, None

    def _pix2windowUnits(self, win_handle, pos):
        win = self._iohub_server._psychopy_windows.get(win_handle)
        win_units = win['units']
        monitor = win['monitor']
        pos = np.asarray(pos)
        if win_units == 'pix':
            return pos
        elif win_units == 'norm':
            return pos * 2.0 / win['size']
        elif win_units == 'cm':
            if monitor:
                return pix2cm(pos, monitor['monitor'])
            else:
                # should raise exception?
                print2err("iohub Mouse error: Window is using units %s but has no Monitor definition." % win_units)
        elif win_units == 'deg':
            if monitor:
                return pix2deg(pos, monitor['monitor'])
            else:
                # should raise exception?
                print2err("iohub Mouse error: Window is using units %s but has no Monitor definition." % win_units)
        elif win_units == 'height':
            return pos / float(win['size'][1])

    def _windowUnits2pix(self, win_handle, pos):
        win = self._iohub_server._psychopy_windows.get(win_handle)
        win_units = win['units']
        monitor = win['monitor']
        pos = np.asarray(pos)
        if win_units == 'pix':
            return pos
        elif win_units == 'norm':
            return pos * win['size'] / 2.0
        elif win_units == 'cm':
            if monitor:
                return cm2pix(pos, monitor['monitor'])
            else:
                # should raise exception?
                print2err("iohub Mouse error: Window is using units %s but has no Monitor definition." % win_units)

        elif win_units == 'deg':
            if monitor:
                return deg2pix(pos, monitor['monitor'])
            else:
                # should raise exception
                print2err("iohub Mouse error: Window is using units %s but has no Monitor definition." % win_units)
        elif win_units == 'height':
            return pos * float(win['size'][1])

    def getDisplayIndex(self):
        """
        Returns the current display index of the ioHub Mouse Device.
        If the display index == the index of the display being used for
        stimulus
        presentation, then mouse position is in the display's coordinate units.
        If the display index != the index of the display being used for
        stimulus
        presentation, then mouse position is in OS system mouse coordinate
        space.

        Args:
            None
        Returns:
            (int): index of the Display the mouse is over. Display index's
            range from 0 to N-1, where N is the number of Display's active
            on the Computer.
        """
        return self._display_index

    def getPositionAndDelta(self, return_display_index=False):
        """Returns a tuple of tuples, being the current position of the ioHub
        Mouse Device as an (x,y) tuple, and the amount the mouse position
        changed the last time it was updated (dx,dy). Mouse Position and Delta
        are in display coordinate units.

        Args:
            None

        Returns:
            tuple: ( (x,y), (dx,dy) ) position of mouse, change in mouse
            position, both in Display coordinate space.

        """
        try:
            self._initialMousePos()
            cpos = self._position
            lpos = self._lastPosition
            if lpos is None:
                lpos = cpos[0], cpos[1]

            change_x = cpos[0] - lpos[0]
            change_y = cpos[1] - lpos[1]
            if return_display_index is True:
                return (cpos, (change_x, change_y), self._display_index)
            return cpos, (change_x, change_y)

        except Exception as e:
            print2err('>>ERROR getPositionAndDelta: ' + str(e))
            printExceptionDetailsToStdErr()
            if return_display_index is True:
                return ((0.0, 0.0), (0.0, 0.0), self._display_index)
            return (0.0, 0.0), (0.0, 0.0)

    def getScroll(self):
        """
        Returns the current vertical scroll value for the mouse. The
        vertical scroll value changes when the
        scroll wheel on a mouse is moved up or down. The vertical scroll
        value is in an arbitrary value space
        ranging for -32648 to +32648. Scroll position is initialize to 0
        when the experiment starts.

        Args:
            None

        Returns:
            int: current vertical scroll value.
        """
        return self._scrollPositionY

    def setScroll(self, s):
        """
        Sets the current vertical scroll value for the mouse. The vertical
        scroll value changes when the
        scroll wheel on a mouse is moved up or down. The vertical scroll
        value is in an
        arbitrary value space ranging for -32648 to +32648. Scroll position
        is initialize to 0 when
        the experiment starts. This method allows you to change the scroll
        value to anywhere in the
        valid value range.

        Args (int):
            The scroll position you want to set the vertical scroll to.
            Should be a number between -32648 to +32648.

        Returns:
            int: current vertical scroll value.
        """
        if isinstance(s, (int, float, complex)):
            self._scrollPositionY = s
        return self._scrollPositionY

    def getDisplayIndexForMousePosition(self, system_mouse_pos):
        return self._display_device._getDisplayIndexForNativePixelPosition(
                system_mouse_pos)

    def _getClippedMousePosForDisplay(self, pixel_pos, display_index):
        drti = self._display_device._getRuntimeInfoByIndex(display_index)
        left, top, right, bottom = drti['bounds']
        mx, my = pixel_pos
        if mx < left:
            mx = left
        elif mx >= right:
            mx = right - 1

        if my < top:
            my = top
        elif my >= bottom:
            my = bottom - 1
        return mx, my

    def _validateMousePosForActiveDisplay(self, pixel_pos):
        left, top, right, bottom = self._display_device.getBounds()
        mx, my = pixel_pos
        mousePositionNeedsUpdate = False

        if mx < left:
            mx = left
            mousePositionNeedsUpdate = True
        elif mx >= right:
            mx = right - 1
            mousePositionNeedsUpdate = True

        if my < top:
            my = top
            mousePositionNeedsUpdate = True
        elif my >= bottom:
            my = bottom - 1
            mousePositionNeedsUpdate = True

        if mousePositionNeedsUpdate:
            return mx, my

        return True

    def _isPixPosWithinDisplay(self, pixel_pos, display_index):
        d = self._display_device._getDisplayIndexForNativePixelPosition(pixel_pos)
        return d == display_index


    def _nativeSetMousePos(self, px, py):
        print2err(
                'ERROR: _nativeSetMousePos must be overwritten by OS '
                'dependent implementation')

if Computer.platform == 'win32':
    from .win32 import Mouse

elif Computer.platform.startswith('linux'):
    from .linux2 import Mouse

elif Computer.platform == 'darwin':
    from .darwin import Mouse

############# OS Independent Mouse Event Classes ####################

from .. import DeviceEvent


class MouseInputEvent(DeviceEvent):
    """The MouseInputEvent is an abstract class that is the parent of all
    MouseInputEvent types that are supported in the ioHub.

    Mouse position is mapped to the coordinate space defined in the
    ioHub configuration file for the Display.

    """
    PARENT_DEVICE = Mouse
    EVENT_TYPE_STRING = 'MOUSE_INPUT'
    EVENT_TYPE_ID = EventConstants.MOUSE_INPUT
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    _newDataTypes = [
        # gives the display index that the mouse was over for the event.
        ('display_id', np.uint8),
        # 1 if button is pressed, 0 if button is released
        ('button_state', np.uint8),
        # 1, 2,or 4, representing left, right, and middle buttons
        ('button_id', np.uint8),
        # sum of currently active button int values
        ('pressed_buttons', np.uint8),

        # x position of the position when the event occurred
        ('x_position', np.float64),
        # y position of the position when the event occurred
        ('y_position', np.float64),

        # horizontal scroll wheel position change when the event occurred (OS X
        # only)
        ('scroll_dx', np.int8),
        # horizontal scroll wheel abs. position when the event occurred (OS X
        # only)
        ('scroll_x', np.int16),
        # vertical scroll wheel position change when the event occurred
        ('scroll_dy', np.int8),
        # vertical scroll wheel abs. position when the event occurred
        ('scroll_y', np.int16),

        # indicates what modifier keys were active when the mouse event
        # occurred.
        ('modifiers', np.uint32),

        # window ID that the mouse was over when the event occurred
        ('window_id', np.uint64)
        # (window does not need to have focus)
    ]

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: The id of the display that the mouse was over when the event
        #: occurred.
        #: Only supported on Windows at this time. Always 0 on other OS's.
        self.display_id = None

        #: 1 if button is pressed, 0 if button is released
        self.button_state = None

        #: MouseConstants.MOUSE_BUTTON_LEFT, MouseConstants.MOUSE_BUTTON_RIGHT
        #: and MouseConstants.MOUSE_BUTTON_MIDDLE are int constants
        #: representing left, right, and middle buttons of the mouse.
        self.button_id = None

        #: 'All' currently pressed button id's logically OR'ed together.
        self.pressed_buttons = None

        #: x position of the Mouse when the event occurred; in display
        #: coordinate space.
        self.x_position = None

        #: y position of the Mouse when the event occurred; in display
        #: coordinate space.
        self.y_position = None

        #: Horizontal scroll wheel position change when the event occurred.
        #: OS X Only. Always 0 on other OS's.
        self.scroll_dx = None

        #: Horizontal scroll wheel absolute position when the event occurred.
        #: OS X Only. Always 0 on other OS's.
        self.scroll_x = None

        #: Vertical scroll wheel position change when the event occurred.
        self.scroll_dy = None

        #: Vertical scroll wheel absolute position when the event occurred.
        self.scroll_y = None

        #: List of the modifiers that were active when the mouse event
        #: occurred,
        #: provided in online events as a list of the modifier constant labels
        #: specified in iohub.ModifierConstants
        #: list: Empty if no modifiers are pressed, otherwise each element is
        #  the string name of a modifier constant.
        self.modifiers = 0

        #: Window handle reference that the mouse was over when the event
        #: occurred
        #: (window does not need to have focus)
        self.window_id = None

        DeviceEvent.__init__(self, *args, **kwargs)

    @classmethod
    def _convertFields(cls, event_value_list):
        modifier_value_index = cls.CLASS_ATTRIBUTE_NAMES.index('modifiers')
        event_value_list[
            modifier_value_index] = KeyboardConstants._modifierCodes2Labels(
                event_value_list[modifier_value_index])

    @classmethod
    def createEventAsDict(cls, values):
        cls._convertFields(values)
        return dict(zip(cls.CLASS_ATTRIBUTE_NAMES, values))

    # noinspection PyUnresolvedReferences
    @classmethod
    def createEventAsNamedTuple(cls, valueList):
        cls._convertFields(valueList)
        return cls.namedTupleClass(*valueList)


class MouseMoveEvent(MouseInputEvent):
    """MouseMoveEvent's occur when the mouse position changes. Mouse position
    is mapped to the coordinate space defined in the ioHub configuration file
    for the Display.

    Event Type ID: EventConstants.MOUSE_MOVE

    Event Type String: 'MOUSE_MOVE'

    """
    EVENT_TYPE_STRING = 'MOUSE_MOVE'
    EVENT_TYPE_ID = EventConstants.MOUSE_MOVE
    IOHUB_DATA_TABLE = MouseInputEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        MouseInputEvent.__init__(self, *args, **kwargs)


class MouseDragEvent(MouseMoveEvent):
    """MouseDragEvents occur when the mouse position changes and at least one
    mouse button is pressed. Mouse position is mapped to the coordinate space
    defined in the ioHub configuration file for the Display.

    Event Type ID: EventConstants.MOUSE_DRAG

    Event Type String: 'MOUSE_DRAG'

    """
    EVENT_TYPE_STRING = 'MOUSE_DRAG'
    EVENT_TYPE_ID = EventConstants.MOUSE_DRAG
    IOHUB_DATA_TABLE = MouseMoveEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        MouseMoveEvent.__init__(self, *args, **kwargs)


class MouseScrollEvent(MouseInputEvent):
    """MouseScrollEvent's are generated when the scroll wheel on the Mouse
    Device (if it has one) is moved. Vertical scrolling is supported on all
    operating systems, horizontal scrolling is only supported on OS X.

    Each MouseScrollEvent provides the number of units the wheel was turned
    in each supported dimension, as well as the absolute scroll value for
    of each supported dimension.

    Event Type ID: EventConstants.MOUSE_SCROLL

    Event Type String: 'MOUSE_SCROLL'

    """
    EVENT_TYPE_STRING = 'MOUSE_SCROLL'
    EVENT_TYPE_ID = EventConstants.MOUSE_SCROLL
    IOHUB_DATA_TABLE = MouseInputEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        """

        :rtype : MouseScrollEvent
        :param args:
        :param kwargs:
        """
        MouseInputEvent.__init__(self, *args, **kwargs)


class MouseButtonEvent(MouseInputEvent):
    EVENT_TYPE_STRING = 'MOUSE_BUTTON'
    EVENT_TYPE_ID = EventConstants.MOUSE_BUTTON
    IOHUB_DATA_TABLE = MouseInputEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        """

        :rtype : MouseButtonEvent
        :param args:
        :param kwargs:
        """
        MouseInputEvent.__init__(self, *args, **kwargs)


class MouseButtonPressEvent(MouseButtonEvent):
    """MouseButtonPressEvent's are created when a button on the mouse is
    pressed. The button_state of the event will equal
    MouseConstants.MOUSE_BUTTON_STATE_PRESSED, and the button that was pressed
    (button_id) will be MouseConstants.MOUSE_BUTTON_LEFT,
    MouseConstants.MOUSE_BUTTON_RIGHT, or MouseConstants.MOUSE_BUTTON_MIDDLE,
    assuming you have a 3 button mouse.

    To get the current state of all three buttons on the Mouse Device,
    the pressed_buttons attribute can be read, which tracks the state of all
    three
    mouse buttons as an int that is equal to the sum of any pressed button id's
    ( MouseConstants.MOUSE_BUTTON_LEFT,  MouseConstants.MOUSE_BUTTON_RIGHT, or
    MouseConstants.MOUSE_BUTTON_MIDDLE ).

    To tell if a given mouse button was depressed when the event occurred,
    regardless of which
    button triggered the event, you can use the following::

        isButtonPressed = event.pressed_buttons &
        MouseConstants.MOUSE_BUTTON_xxx == MouseConstants.MOUSE_BUTTON_xxx

    where xxx is LEFT, RIGHT, or MIDDLE.

    For example, if at the time of the event both the left and right mouse
    buttons
    were in a pressed state::

        buttonToCheck=MouseConstants.MOUSE_BUTTON_RIGHT
        isButtonPressed = event.pressed_buttons & buttonToCheck ==
        buttonToCheck

        print isButtonPressed

        >> True

        buttonToCheck=MouseConstants.MOUSE_BUTTON_LEFT
        isButtonPressed = event.pressed_buttons & buttonToCheck ==
        buttonToCheck

        print isButtonPressed

        >> True

        buttonToCheck=MouseConstants.MOUSE_BUTTON_MIDDLE
        isButtonPressed = event.pressed_buttons & buttonToCheck ==
        buttonToCheck

        print isButtonPressed

        >> False

    Event Type ID: EventConstants.MOUSE_BUTTON_PRESS

    Event Type String: 'MOUSE_BUTTON_PRESS'

    """
    EVENT_TYPE_STRING = 'MOUSE_BUTTON_PRESS'
    EVENT_TYPE_ID = EventConstants.MOUSE_BUTTON_PRESS
    IOHUB_DATA_TABLE = MouseInputEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        MouseButtonEvent.__init__(self, *args, **kwargs)


class MouseButtonReleaseEvent(MouseButtonEvent):
    """MouseButtonUpEvent's are created when a button on the mouse is released.

    The button_state of the event will equal
    MouseConstants.MOUSE_BUTTON_STATE_RELEASED,
    and the button that was pressed (button_id) will be
    MouseConstants.MOUSE_BUTTON_LEFT,
    MouseConstants.MOUSE_BUTTON_RIGHT, or MouseConstants.MOUSE_BUTTON_MIDDLE,
    assuming you have a 3 button mouse.

    Event Type ID: EventConstants.MOUSE_BUTTON_RELEASE

    Event Type String: 'MOUSE_BUTTON_RELEASE'

    """
    EVENT_TYPE_STRING = 'MOUSE_BUTTON_RELEASE'
    EVENT_TYPE_ID = EventConstants.MOUSE_BUTTON_RELEASE
    IOHUB_DATA_TABLE = MouseInputEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        MouseButtonEvent.__init__(self, *args, **kwargs)


class MouseMultiClickEvent(MouseButtonEvent):
    """MouseMultiClickEvent's are created when you rapidly press and release a
    mouse button two or more times. This event may never get triggered if your
    OS does not support it. The button that was multi clicked (button_id) will
    be MouseConstants.MOUSE_BUTTON_LEFT, MouseConstants.MOUSE_BUTTON_RIGHT, or
    MouseConstants.MOUSE_BUTTON_MIDDLE, assuming you have a 3 button mouse.

    Event Type ID: EventConstants.MOUSE_MULTI_CLICK

    Event Type String: 'MOUSE_MULTI_CLICK'

    """
    EVENT_TYPE_STRING = 'MOUSE_MULTI_CLICK'
    EVENT_TYPE_ID = EventConstants.MOUSE_MULTI_CLICK
    IOHUB_DATA_TABLE = MouseInputEvent.IOHUB_DATA_TABLE
    __slots__ = []

    def __init__(self, *args, **kwargs):
        MouseButtonEvent.__init__(self, *args, **kwargs)



from psychopy import core, logging, hardware
import pytest

logging.console.setLevel(logging.DEBUG)

def test_PR655():
    pr655 = hardware.findPhotometer(device='PR655')
    if pr655 is None:
        logging.warning('no device found')
    else:
        print(('type:', pr655.type))
        print(('SN:', pr655.getDeviceSN()))
        #on linux we do actually find a device that returns 'D'
        if pr655.type=='D':
            pytest.skip()
        pr655.measure()
        print(('lum', pr655.lastLum))
        print(('uv',pr655.lastUV))
        print(('xy', pr655.lastXY))
        print(('tristim', pr655.lastTristim))
        nm, spec = pr655.getLastSpectrum()
        print(('nm', nm))
        print(('spec', spec))
        print(('temperature', pr655.lastColorTemp))

    print('DONE')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import sys
from .errors import print2err

# pylint: disable=too-many-lines

class Constants:
    UNDEFINED = 0
    _keys = None
    _names = None
    _classes = None
    _initialized = False

    @classmethod
    def getName(cls, cid):
        """Return the constant's name given a valid constant id.

        Args:
            cid (int): Constant's id representation.

        Returns:
            str: The name for the given constant id.

        """
        return cls._names.get(cid, cls._names[cls.UNDEFINED])

    @classmethod
    def getID(cls, name):
        """Return the constant's id given a valid constant name string.

        Args:
            name (str): Constant's name representation.

        Returns:
            int: The id for the given constant name.

        """
        return cls._names.get(name, None)

    @classmethod
    def getClass(cls, cid):
        """Return the constant's ioHub CLass Name given constant id. If no
        class is associated with the specified constant value, None is
        returned.

        Args:
            cid (int): Constant's id representation.

        Returns:
            class: The ioHub class for the constant id provided.

        """
        return cls._classes.get(cid, None)

    @classmethod
    def initialize(cls, starting_index=1):
        if cls._initialized:
            return
        for i, a in enumerate(dir(cls)):
            if ((a[0] != '_') and (not callable(getattr(cls, a)))
                    and (getattr(cls, a) < 0)):
                setattr(cls, a, i + starting_index)

        cls._names = dict([(getattr(cls, a), a) for a in dir(cls) if (
            (a[0] != '_') and (not callable(getattr(cls, a))))])
        cls._keys = list(cls._names.keys())
        cls._names.update(dict([(v, k)
                                for k, v in cls._names.items()]))
        cls._initialized = True

    @classmethod
    def getConstants(cls):
        return cls._names


class EventConstants(Constants):
    """
    EventConstants contains ioHub Device Event type constants, with
    methods to convert between the different associated constant value
    types for a given event type:

    * int constant
    * str constant
    * event class associated with a constant

    Access event constants using the EventConstants class attributes & methods.
    """
    KEYBOARD_INPUT = 20
    #: Constant for a Keyboard Press Event.
    KEYBOARD_PRESS = 22
    #: Constant for a Keyboard Release Event.
    KEYBOARD_RELEASE = 23

    MOUSE_INPUT = 30
    MOUSE_BUTTON = 31
    #: Constant for a Mouse Button Press Event.
    MOUSE_BUTTON_PRESS = 32

    #: Constant for a Mouse Button Release Event.
    MOUSE_BUTTON_RELEASE = 33

    #: Constant for a Mouse Double Click Event.
    #: Deprecated for MOUSE_MULTI_CLICK in 0.6RC1
    MOUSE_DOUBLE_CLICK = 34

    #: Constant for a Mouse Multiple Click Event.
    MOUSE_MULTI_CLICK = 34

    #: Constant for a Mouse Scroll Wheel Event.
    MOUSE_SCROLL = 35

    #: Constant for a Mouse Move Event.
    MOUSE_MOVE = 36

    #: Constant for a Mouse Drag Event.
    MOUSE_DRAG = 37

    TOUCH = 40
    #: Constant for a Touch motion Event.
    TOUCH_MOVE = 41

    #: Constant for an initial Touch press Event.
    TOUCH_PRESS = 42

    #: Constant for a Touch release Event.
    TOUCH_RELEASE = 43

    EYETRACKER = 50
    #: Constant for an Eye Tracker Monocular Sample Event.
    MONOCULAR_EYE_SAMPLE = 51

    #: Constant for an Eye Tracker Binocular Sample Event.
    BINOCULAR_EYE_SAMPLE = 52

    #: Constant for an Eye Tracker Fixation Start Event.
    FIXATION_START = 53

    #: Constant for an Eye Tracker Fixation End Event.
    FIXATION_END = 54

    #: Constant for an Eye Tracker Saccade Start Event.
    SACCADE_START = 55

    #: Constant for an Eye Tracker Saccade End Event.
    SACCADE_END = 56

    #: Constant for an Eye Tracker Blink Start Event.
    BLINK_START = 57

    #: Constant for an Eye Tracker Blink End Event.
    BLINK_END = 58

    #: Constant for a Basic Eye Tracker Sample Event.
    EYE_SAMPLE = 59

    #: Constant for a GazePoint specific Sample Event.
    GAZEPOINT_SAMPLE = 60

    #: Constant for a Gamepad Event.
    GAMEPAD_STATE_CHANGE = 81
    GAMEPAD_DISCONNECT = 82

    WINTAB_INPUT = 90
    #: Constant for a WinTab Pen Sample Event.
    WINTAB_SAMPLE = 91

    #: Constant for a WinTab Pen Region Entered Event.
    WINTAB_ENTER_REGION = 92

    #: Constant for a WinTab Pen Region Leave Event.
    WINTAB_LEAVE_REGION = 93

    #: Constant for MCU digital input Event.
    DIGITAL_INPUT = 101

    #: Constant for MCU analog input Event.
    ANALOG_INPUT = 102

    #: Constant for MCU analog threshold state change Event.
    THRESHOLD = 103

    #: Constant for a general purpose Serial Rx Event.
    SERIAL_INPUT = 105

    #: Constant for a serial event due to a rx stream byte value diff.
    SERIAL_BYTE_CHANGE = 106

    #: Constant for a PST Box event due to a button state change.
    PSTBOX_BUTTON = 107

    #: Constant for an Eight Channel Analog Input Sample Event.
    MULTI_CHANNEL_ANALOG_INPUT = 122

    #: Constant for an Experiment Message Event.
    MESSAGE = 151

    #: Constant for an Experiment Log Event.
    LOG = 152

    def __init__(self):
        pass

    @classmethod
    def addClassMappings(cls, device_event_ids, event_classes):
        if cls._classes is None:
            cls._classes = {}

        for event_id in device_event_ids:
            event_constant_string = cls.getName(event_id)
            event_class = None
            for event_class in event_classes.values():
                if event_class.EVENT_TYPE_ID == event_id:
                    cls._classes[event_id] = event_class
                    cls._classes[event_class] = event_id
                    break

            if event_id not in cls._classes:
                print2err('\t*** ERROR ADDING EVENT CLASS MAPPING:',
                          ' Could not find class: ',
                          event_constant_string, ' = ', event_id)

EventConstants.initialize()

EYE_SAMPLE_TYPES = [EventConstants.EYE_SAMPLE, EventConstants.MONOCULAR_EYE_SAMPLE,
                    EventConstants.BINOCULAR_EYE_SAMPLE, EventConstants.GAZEPOINT_SAMPLE]

class DeviceConstants(Constants):
    """
    DeviceConstants contains the ioHub Device type constants, with
    methods to convert between the different associated constant value
    types for a given device type::

    * int constant
    * str constant
    * device class associated with a constant

    Access device constants using the DeviceConstants class attributes
    & methods.
    """
    #: Constant for a Device Type not currently categorized.
    OTHER = 1

    #: Constant for a Keyboard Device.
    KEYBOARD = 20

    #: Constant for a Mouse Device.
    MOUSE = 30

    #: Constant for a Touch Device.
    TOUCH = 40

    #: Constant for an EyeTracker Device.
    EYETRACKER = 50

    #: Constant for a Network Device
    EVENTPUBLISHER = 61

    #: Constant for a Network Device
    REMOTEEVENTSUBSCRIBER = 62

    XINPUT = 70

    #: Constant for Gamepad Device.
    GAMEPAD = 80

    #: Constant for Tablet Device that uses WinTab DLL.
    WINTAB = 90

    #: Constant for a MCU Device.
    MCU = 100

    #: Constant for a General Purpose Serial Interface Device.
    SERIAL = 110

    #: Constant for PST Serial Response Box
    PSTBOX = 111

    #: Constant for an AnalogInput Device.
    ANALOGINPUT = 120

    #: Constant for an Experiment Device.
    EXPERIMENT = 150

    #: Constant for a Display Device.
    DISPLAY = 190

    #: Constant for a Computer Device.
    COMPUTER = 200

    def __init__(self):
        pass

    @classmethod
    def addClassMapping(cls, device_class):
        if cls._classes is None:
            cls._classes = {}

        device_constant_string = device_class.__name__.upper()
        device_id = getattr(cls, device_constant_string)
        cls._classes[device_id] = device_class
        cls._classes[device_class] = device_id

DeviceConstants.initialize()


class MouseConstants(Constants):
    """MouseConstants provides access to ioHub Mouse Device specific
    constants."""
    #: Constant representing that no Mouse buttons are pressed.
    MOUSE_BUTTON_NONE = 0

    #: Constant representing that the left Mouse button is pressed.
    MOUSE_BUTTON_LEFT = 1

    #: Constant representing that the right Mouse button is pressed.
    MOUSE_BUTTON_RIGHT = 2

    #: Constant representing that the middle Mouse button is pressed.
    MOUSE_BUTTON_MIDDLE = 4

    #: Constant representing a mouse button is in a released state.
    MOUSE_BUTTON_STATE_RELEASED = 10

    #: Constant representing a mouse button is in a pressed state.
    MOUSE_BUTTON_STATE_PRESSED = 11

    #: Constant representing a mouse is in a multiple click state.
    MOUSE_BUTTON_STATE_MULTI_CLICK = 12

    MOUSE_BUTTON_4 = 8
    MOUSE_BUTTON_5 = 16
    MOUSE_BUTTON_6 = 32
    MOUSE_BUTTON_7 = 64
    MOUSE_BUTTON_8 = 128
    MOUSE_BUTTON_9 = 256

    def __init__(self):
        pass

MouseConstants.initialize()

# pylint: disable=protected-access

if sys.platform == 'win32':
    class AsciiConstants(Constants):
        # Mainly from the pyHook lookup Table, some from Pyglet
        BACKSPACE = 0x08
        TAB = 0x09
        LINEFEED = 0x0A
        CLEAR = 0x0B
        RETURN = 0x0D
        SYSREQ = 0x15
        ESCAPE = 0x1B

        SPACE = 0x20
        EXCLAMATION = 0x21
        DOUBLEQUOTE = 0x22
        POUND = 0x23
        DOLLAR = 0x24
        PERCENT = 0x25
        AMPERSAND = 0x26
        APOSTROPHE = 0x27
        PARENLEFT = 0x28
        PARENRIGHT = 0x29
        ASTERISK = 0x2A
        PLUS = 0x2B
        COMMA = 0x2C
        MINUS = 0x2D
        PERIOD = 0x2E
        SLASH = 0x2F

        n0_ = 0x30
        n1_ = 0x31
        n2_ = 0x32
        n3_ = 0x33
        n4_ = 0x34
        n5_ = 0x35
        n6_ = 0x36
        n7_ = 0x37
        n8_ = 0x38
        n9_ = 0x39

        COLON = 0x3A
        SEMICOLON = 0x3B
        LESS = 0x3C
        EQUAL = 0x3D
        GREATER = 0x3E
        QUESTION = 0x3F
        AT = 0x40

        A = 0x41
        B = 0x42
        C = 0x43
        D = 0x44
        E = 0x45
        F = 0x46
        G = 0x47
        H = 0x48
        I = 0x49
        J = 0x4A
        K = 0x4B
        L = 0x4C
        M = 0x4D
        N = 0x4E
        O = 0x4F
        P = 0x50
        Q = 0x51
        R = 0x52
        S = 0x53
        T = 0x54
        U = 0x55
        V = 0x56
        W = 0x57
        X = 0x58
        Y = 0x59
        Z = 0x5A

        BRACKETLEFT = 0x5B
        BACKSLASH = 0x5C
        BRACKETRIGHT = 0x5D
        ASCIICIRCUM = 0x5E
        UNDERSCORE = 0x5F
        GRAVE = 0x60

        a = 0x61
        b = 0x62
        c = 0x63
        d = 0x64
        e = 0x65
        f = 0x66
        g = 0x67
        h = 0x68
        i = 0x69
        j = 0x6A
        k = 0x6B
        l = 0x6C
        m = 0x6D
        n = 0x6E
        o = 0x6F
        p = 0x70
        q = 0x71
        r = 0x72
        s = 0x73
        t = 0x74
        u = 0x75
        v = 0x76
        w = 0x77
        x = 0x78
        y = 0x79
        z = 0x7A

        BRACELEFT = 0x7B
        BAR = 0x7C
        BRACERIGHT = 0x7D
        ASCIITILDE = 0x7E

        @classmethod
        def getName(cls, cid):
            return cls._names.get(cid, None)

    AsciiConstants.initialize()

    class VirtualKeyCodes(Constants):
        # Mainly from the pyHook lookup Table, some from Pyglet
        VK_CANCEL = 0x03
        VK_BACKSPACE = 0x08
        VK_TAB = 0x09
        VK_CLEAR = 0x0C
        VK_RETURN = 0x0D

        VK_SHIFT = 0x10
        VK_CONTROL = 0x11
        VK_MENU = 0x12
        VK_PAUSE = 0x13
        VK_capslock = 0x14
        VK_CAPITAL = 0x14
        VK_CAPS_LOCK = 0x14
        VK_HANGUL = 0x15
        VK_JUNJA = 0x17
        VK_FINAL = 0x18
        VK_HANJA = 0x19
        VK_ESCAPE = 0x1B
        VK_CONVERT = 0x1C
        VK_NONCONVERT = 0x1D
        VK_ACCEPT = 0x1E
        VK_MODECHANGE = 0x1F

        VK_SPACE = 0x20
        VK_PAGEUP = 0x21
        VK_PAGEDOWN = 0x22
        VK_END = 0x23
        VK_HOME = 0x24
        VK_LEFT = 0x25
        VK_UP = 0x26
        VK_RIGHT = 0x27
        VK_DOWN = 0x28
        VK_SELECT = 0x29
        VK_PRINT = 0x2A
        VK_EXECUTE = 0x2B
        VK_PRINT_SCREEN = 0x2C
        VK_INSERT = 0x2D
        VK_DELETE = 0x2E
        VK_HELP = 0x2F

        VK_LWIN = 0x5B
        VK_RWIN = 0x5C
        VK_APPS = 0x5D
        VK_lcmd = 0x5B
        VK_rcmd = 0x5C
        VK_menu = 0x5D

        VK_NUMPAD0 = 0x60
        VK_NUMPAD1 = 0x61
        VK_NUMPAD2 = 0x62
        VK_NUMPAD3 = 0x63
        VK_NUMPAD4 = 0x64
        VK_NUMPAD5 = 0x65
        VK_NUMPAD6 = 0x66
        VK_NUMPAD7 = 0x67
        VK_NUMPAD8 = 0x68
        VK_NUMPAD9 = 0x69
        VK_NUMPADMULTIPLY = 0x6A
        VK_NUMPADADD = 0x6B
        VK_SEPARATOR = 0x6C
        VK_NUMPADSUBTRACT = 0x6D
        VK_NUMPADDECIMAL = 0x6E
        VK_NUMPADDIVIDE = 0x6F

        VK_F1 = 0x70
        VK_F2 = 0x71
        VK_F3 = 0x72
        VK_F4 = 0x73
        VK_F5 = 0x74
        VK_F6 = 0x75
        VK_F7 = 0x76
        VK_F8 = 0x77
        VK_F9 = 0x78
        VK_F10 = 0x79
        VK_F11 = 0x7A
        VK_F12 = 0x7B
        VK_F13 = 0x7C
        VK_F14 = 0x7D
        VK_F15 = 0x7E
        VK_F16 = 0x7F
        VK_F17 = 0x80
        VK_F18 = 0x81
        VK_F19 = 0x82
        VK_F20 = 0x83
        VK_F21 = 0x84
        VK_F22 = 0x85
        VK_F23 = 0x86
        VK_F24 = 0x87

        VK_NUM_LOCK = 0x90
        VK_SCROLL = 0x91
        VK_LSHIFT = 0xA0
        VK_RSHIFT = 0xA1
        VK_LCONTROL = 0xA2
        VK_RCONTROL = 0xA3
        VK_LMENU = 0xA4
        VK_RMENU = 0xA5

        VK_numlock = 0x90
        VK_scrolllock = 0x91
        VK_lshift = 0xA0
        VK_rshift = 0xA1
        VK_lctrl = 0xA2
        VK_rctrl = 0xA3
        VK_lalt = 0xA4
        VK_ralt = 0xA5

        VK_BROWSER_BACK = 0xA6
        VK_BROWSER_FORWARD = 0xA7
        VK_BROWSER_REFRESH = 0xA8
        VK_BROWSER_STOP = 0xA9
        VK_BROWSER_SEARCH = 0xAA
        VK_BROWSER_FAVORITES = 0xAB
        VK_BROWSER_HOME = 0xAC
        VK_VOLUME_MUTE = 0xAD
        VK_VOLUME_DOWN = 0xAE
        VK_VOLUME_UP = 0xAF

        VK_MEDIA_NEXT_TRACK = 0xB0
        VK_MEDIA_PREV_TRACK = 0xB1
        VK_MEDIA_STOP = 0xB2
        VK_MEDIA_PLAY_PAUSE = 0xB3
        VK_LAUNCH_MAIL = 0xB4
        VK_LAUNCH_MEDIA_SELECT = 0xB5
        VK_LAUNCH_APP1 = 0xB6
        VK_LAUNCH_APP2 = 0xB7

        VK_PROCESSKEY = 0xE5
        VK_PACKET = 0xE7
        VK_ATTN = 0xF6
        VK_CRSEL = 0xF7
        VK_EXSEL = 0xF8
        VK_EREOF = 0xF9
        VK_PLAY = 0xFA
        VK_ZOOM = 0xFB
        VK_NONAME = 0xFC
        VK_PA1 = 0xFD
        VK_OEM_CLEAR = 0xFE

        @classmethod
        def getName(cls, cid):
            return cls._names.get(cid, None)

    VirtualKeyCodes.initialize()

elif sys.platform.startswith('linux'):
    class VirtualKeyCodes(Constants):
        @classmethod
        def getName(cls, cid):
            return cls._names.get(cid, None)
    VirtualKeyCodes.initialize()

elif sys.platform == 'darwin':
    class AnsiKeyCodes(Constants):
        ANSI_Equal = 0x18
        ANSI_Minus = 0x1B
        ANSI_RightBracket = 0x1E
        ANSI_LeftBracket = 0x21
        ANSI_Quote = 0x27
        ANSI_Semicolon = 0x29
        ANSI_Backslash = 0x2A
        ANSI_Comma = 0x2B
        ANSI_Slash = 0x2C
        ANSI_Period = 0x2F
        ANSI_Grave = 0x32
        ANSI_KeypadDecimal = 0x41
        ANSI_KeypadMultiply = 0x43
        ANSI_KeypadPlus = 0x45
        ANSI_KeypadClear = 0x47
        ANSI_KeypadDivide = 0x4B
        ANSI_KeypadEnter = 0x4C
        ANSI_KeypadMinus = 0x4E
        ANSI_KeypadEquals = 0x51
        ANSI_Keypad0 = 0x52
        ANSI_Keypad1 = 0x53
        ANSI_Keypad2 = 0x54
        ANSI_Keypad3 = 0x55
        ANSI_Keypad4 = 0x56
        ANSI_Keypad5 = 0x57
        ANSI_Keypad6 = 0x58
        ANSI_Keypad7 = 0x59
        ANSI_Keypad8 = 0x5B
        ANSI_Keypad9 = 0x5C
        @classmethod
        def getName(cls, cid):
            return cls._names.get(cid, None)

    AnsiKeyCodes.initialize()
    AnsiKeyCodes._keys.remove(AnsiKeyCodes.getID('UNDEFINED'))

    class UnicodeChars(Constants):
        VK_RETURN = 0x0003
        RETURN = 0x000D
        VK_DELETE = 0x007F  # "Delete"
        TAB = 0x0009  # "Tab"
        ESCAPE = 0x001b  # "Escape"
        UP = 0xf700  # "Up"
        DOWN = 0xF701  # "Down"
        LEFT = 0xF702  # "Left"
        RIGHT = 0xF703  # "Right"
        F1 = 0xF704  # "F1"
        F2 = 0xF705  # "F2"
        F3 = 0xF706  # "F3"
        F4 = 0xF707  # "F4"
        F5 = 0xF708  # "F5"
        F6 = 0xF709  # "F6"
        F7 = 0xF70A  # "F7"
        F8 = 0xF70B  # "F8"
        F9 = 0xF70C  # "F9"
        F10 = 0xF70D  # "F10"
        F11 = 0xF70E  # "F11"
        F12 = 0xF70F  # "F12"
        F13 = 0xF710  # "F13"
        F14 = 0xF711  # "F14"
        F15 = 0xF712  # "F15"
        F16 = 0xF713  # "F16"
        F17 = 0xF714  # "F17"
        F18 = 0xF715  # "F18"
        F19 = 0xF716  # "F19"
        F20 = 0xF717  # "F20"
        F21 = 0xF718  # "F21"
        F22 = 0xF719  # "F22"
        F23 = 0xF71A  # "F23"
        F24 = 0xF71B  # "F24"
        F25 = 0xF71C  # "F25"
        F26 = 0xF71D  # "F26"
        F27 = 0xF71E  # "F27"
        F28 = 0xF71F  # "F28"
        F29 = 0xF720  # "F29"
        F30 = 0xF721  # "F30"
        F31 = 0xF722  # "F31"
        F32 = 0xF723  # "F32"
        F33 = 0xF724  # "F33"
        F34 = 0xF725  # "F34"
        F35 = 0xF726  # "F35"
        INSERT = 0xF727  # "Insert"
        DELETE = 0xF728  # "Delete"
        HOME = 0xF729  # "Home"
        BEGIN = 0xF72A  # "Begin"
        END = 0xF72B  # "End"
        PAGE_UP = 0xF72C  # "PageUp"
        PAGE_DOWN = 0xF72D  # "PageDown"
        PRINT_SCREEN = 0xF72E  # "PrintScreen"
        SCROLL_LOCK = 0xF72F  # "ScrollLock"
        PAUSE = 0xF730  # "Pause"
        SYSREQ = 0xF731  # "SysReq"
        BREAK = 0xF732  # "Break"
        RESET = 0xF733  # "Reset"
        STOP = 0xF734  # "Stop"
        MENU = 0xF735  # "Menu"
        VK_MENU = 0x0010  # Menu key on non-apple US keyboards
        USER = 0xF736  # "User"
        SYSTEM = 0xF737  # "System"
        PRINT = 0xF738  # "Print"
        CLEAR_LINE = 0xF739  # "ClearLine"
        CLEAR = 0xF73A  # "ClearDisplay"
        INSERT_LINE = 0xF73B  # "InsertLine"
        DELETE_LINE = 0xF73C  # "DeleteLine"
        INSERT_CHAR = 0xF73D  # "InsertChar"
        DELETE_CHAR = 0xF73E  # "DeleteChar"
        PREV = 0xF73F  # "Prev"
        NEXT = 0xF740  # "Next"
        SELECT = 0xF741  # "Select"
        EXECUTE = 0xF742  # "Execute"
        UNDO = 0xF743  # "Undo"
        REDO = 0xF744  # "Redo"
        FIND = 0xF745  # "Find"
        HELP = 0xF746  # "Help"
        MODE = 0xF747  # "ModeSwitch"
        SHIFT = 0x21E7  # Unicode UPWARDS WHITE ARROW
        CONTROL = 0x2303  # Unicode UP ARROWHEAD
        OPTION = 0x2325  # Unicode OPTION KEY
        COMMAND = 0x2318  # Unicode PLACE OF INTEREST SIGN
        # Unicode LOWER RIGHT PENCIL; actually pointed left until Mac OS X 10.3
        PENCIL_RIGHT = 0x270E
        # Unicode LOWER LEFT PENCIL; available in Mac OS X 10.3 and later
        PENCIL_LEFT = 0xF802
        CHECK = 0x2713  # Unicode CHECK MARK
        DIAMOND = 0x25C6  # Unicode BLACK DIAMOND
        BULLET = 0x2022  # Unicode BULLET
        APPLE_LOGO = 0xF8FF  # Unicode APPLE LOGO

        @classmethod
        def getName(cls, cid):
            return cls._names.get(cid, None)

    UnicodeChars.initialize()
    UnicodeChars._keys.remove(UnicodeChars.getID('UNDEFINED'))

    class VirtualKeyCodes(Constants):
        F1 = 145  # Keycode on Apple wireless kb
        F2 = 144  # Keycode on Apple wireless kb
        F3 = 160  # Keycode on Apple wireless kb
        F4 = 131  # Keycode on Apple wireless kb
        VK_ISO_SECTION = 0x0A
        VK_JIS_YEN = 0x5D
        VK_JIS_UNDERSCORE = 0x5E
        VK_JIS_KEYPAD_COMMA = 0x5F
        VK_JIS_EISU = 0x66
        VK_JIS_KANA = 0x68
        VK_RETURN = 0x24
        VK_TAB = 0x30
        VK_SPACE = 0x31
        VK_DELETE = 0x33
        VK_ESCAPE = 0x35
        VK_COMMAND = 0x37
        VK_SHIFT = 0x38
        VK_CAPS_LOCK = 0x39
        VK_OPTION = 0x3A
        VK_CONTROL = 0x3B
        VK_SHIFT_RIGHT = 0x3C
        VK_OPTION_RIGHT = 0x3D
        VK_CONTROL_RIGHT = 0x3E
        VK_FUNCTION = 0x3F
        VK_F17 = 0x40
        VK_VOLUME_UP = 0x48
        VK_VOLUME_DOWN = 0x49
        VK_VOLUME_MUTE = 0x4A
        VK_F18 = 0x4F
        VK_F19 = 0x50
        VK_F20 = 0x5A
        VK_F5 = 0x60
        VK_F6 = 0x61
        VK_F7 = 0x62
        VK_F3 = 0x63
        VK_F8 = 0x64
        VK_F9 = 0x65
        VK_F11 = 0x67
        VK_F13 = 0x69
        VK_F16 = 0x6A
        VK_F14 = 0x6B
        VK_F10 = 0x6D
        VK_MENU = 0x6E
        VK_F12 = 0x6F
        VK_F15 = 0x71
        VK_HELP = 0x72
        VK_HOME = 0x73
        VK_PAGE_UP = 0x74
        VK_DEL = 0x75
        VK_F4 = 0x76
        VK_END = 0x77
        VK_F2 = 0x78
        VK_PAGE_DOWN = 0x79
        VK_LEFT = 0x7B
        VK_UP = 0x7E
        VK_RIGHT = 0x7C
        VK_DOWN = 0x7D
        VK_F1 = 0x7A
        KEY_EQUAL = 24
        KEY_MINUS = 27
        KEY_RIGHT_SQUARE_BRACKET = 30
        KEY_LEFT_SQUARE_BRACKET = 33
        KEY_RETURN = 36
        KEY_SINGLE_QUOTE = 39
        KEY_SEMICOLAN = 41
        KEY_BACKSLASH = 42
        KEY_COMMA = 43
        KEY_FORWARD_SLASH = 44
        KEY_PERIOD = 47
        KEY_TAB = 48
        KEY_SPACE = 49
        KEY_LEFT_SINGLE_QUOTE = 50
        KEY_DELETE = 51
        KEY_ENTER = 52
        KEY_ESCAPE = 53
        KEYPAD_PERIOD = 65
        KEYPAD_MULTIPLY = 67
        KEYPAD_PLUS = 69
        KEYPAD_CLEAR = 71
        KEYPAD_DIVIDE = 75
        KEYPAD_ENTER = 76   # numberpad on full kbd
        KEYPAD_EQUALS = 78
        KEYPAD_EQUAL = 81
        KEYPAD_0 = 82
        KEYPAD_1 = 83
        KEYPAD_2 = 84
        KEYPAD_3 = 85
        KEYPAD_4 = 86
        KEYPAD_5 = 87
        KEYPAD_6 = 88
        KEYPAD_7 = 89
        KEYPAD_8 = 91
        KEYPAD_9 = 92
        KEY_F5 = 96
        KEY_F6 = 97
        KEY_F7 = 98
        KEY_F3 = 99
        KEY_F8 = 100
        KEY_F9 = 101
        KEY_F11 = 103
        KEY_F13 = 105
        KEY_F14 = 107
        KEY_F10 = 109
        KEY_F12 = 111
        KEY_F15 = 113
        KEY_HELP = 114
        KEY_HOME = 115
        KEY_PGUP = 116
        KEY_DELETE = 117
        KEY_F4 = 118
        KEY_END = 119
        KEY_F2 = 120
        KEY_PGDN = 121
        KEY_F1 = 122
        KEY_LEFT = 123
        KEY_RIGHT = 124
        KEY_DOWN = 125
        KEY_UP = 126

        @classmethod
        def getName(cls, cid):
            return cls._names.get(cid, None)

    VirtualKeyCodes.initialize()
    VirtualKeyCodes._keys.remove(VirtualKeyCodes.getID('UNDEFINED'))


class ModifierKeyCodes(Constants):
    _mod_names = [
        'lctrl',
        'rctrl',
        'lshift',
        'rshift',
        'lalt',
        'ralt',
        'lcmd',
        'rcmd',
        'capslock',
        'MOD_SHIFT',
        'MOD_ALT',
        'MOD_CTRL',
        'MOD_CMD',
        'numlock',
        'function',
        'modhelp',
        'scrolllock']
    lctrl = 1
    rctrl = 2
    lshift = 4
    rshift = 8
    lalt = 16
    ralt = 32
    lcmd = 64
    rcmd = 128
    capslock = 256
    MOD_SHIFT = 512
    MOD_ALT = 1024
    MOD_CTRL = 2048
    MOD_CMD = 4096
    numlock = 8192
    function = 16384
    modhelp = 32768
    scrolllock = modhelp * 2
ModifierKeyCodes.initialize()
ModifierKeyCodes._keys.remove(ModifierKeyCodes.getID('UNDEFINED'))


class KeyboardConstants(Constants):
    """Stores internally used mappings between OS and iohub keyboard key
    constants."""

    _virtualKeyCodes = VirtualKeyCodes()
    if sys.platform == 'win32':
        _asciiKeyCodes = AsciiConstants()
    if sys.platform == 'darwin':
        _unicodeChars = UnicodeChars()
        _ansiKeyCodes = AnsiKeyCodes()
    _modifierCodes = ModifierKeyCodes()

    @classmethod
    def getName(cls, cid):
        return cls._names.get(cid, None)

    @classmethod
    def _getKeyName(cls, keyEvent):
        vcode_name = KeyboardConstants._virtualKeyCodes.getName(
            keyEvent.KeyID)
        if vcode_name:
            if vcode_name.startswith('VK_NUMPAD'):
                return 'num_%s' % (vcode_name[9:].lower())
            return vcode_name[3:].lower()
        else:
            phkey = keyEvent.Key.lower()
            if phkey.startswith('numpad'):
                phkey = 'num_%s' % (phkey.Key[6:])
            return phkey

    @classmethod
    def _getKeyNameAndModsForEvent(cls, keyEvent):
        return cls._getKeyName(
            keyEvent), cls.getModifiersForEvent(keyEvent)

    @classmethod
    def getModifiersForEvent(cls, event):
        return cls._modifierCodes2Labels(event.Modifiers)

    @classmethod
    def _modifierCodes2Labels(cls, mods):
        if mods == 0:
            return []

        modconstants = cls._modifierCodes
        modNameList = []
        for k in modconstants._keys:
            mc = modconstants._names[k]
            if mods & k == k:
                modNameList.append(mc)
                mods = mods - k
                if mods == 0:
                    return modNameList
        return modNameList

KeyboardConstants.initialize()


class EyeTrackerConstants(Constants):
    # Sample Filter Levels
    FILTER_LEVEL_OFF = 0
    FILTER_OFF = 0
    FILTER_LEVEL_1 = 1
    FILTER_LEVEL_2 = 2
    FILTER_LEVEL_3 = 3
    FILTER_LEVEL_4 = 4
    FILTER_LEVEL_5 = 5
    FILTER_ON = 9

    # Sample Filter Types
    FILTER_FILE = 10
    FILTER_NET = 11
    FILTER_ONLINE = 11
    FILTER_SERIAL = 12
    FILTER_ANALOG = 13
    FILTER_ALL = 14

    # Eye Type Constants
    LEFT_EYE = 21
    RIGHT_EYE = 22
    SIMULATED_MONOCULAR = 23
    MONOCULAR = 24
    BINOCULAR = 26
    BINOCULAR_AVERAGED = 27
    BINOCULAR_CUSTOM = 28
    SIMULATED_BINOCULAR = 29


    # Calibration / Validation Related Constants

    # Target Point Count
    NO_POINTS = 40
    ONE_POINT = 41
    TWO_POINTS = 42
    THREE_POINTS = 43
    FOUR_POINTS = 44
    FIVE_POINTS = 45
    SEVEN_POINTS = 47
    EIGHT_POINTS = 48
    NINE_POINTS = 49
    THIRTEEN_POINTS = 53
    SIXTEEN_POINTS = 56
    TWENTYFIVE_POINTS = 65
    CUSTOM_POINTS = 69

    # Pattern Dimensionality Types
    CALIBRATION_HORZ_1D = 130
    CALIBRATION_VERT_1D = 131
    CALIBRATION_2D = 132
    CALIBRATION_3D = 133

    # Target Pacing Types
    AUTO_CALIBRATION_PACING = 90
    MANUAL_CALIBRATION_PACING = 91

    # Target Shape Types
    CIRCLE_TARGET = 121
    CROSSHAIR_TARGET = 122
    IMAGE_TARGET = 123
    MOVIE_TARGET = 124

    # System Setup Method Initial State Constants
    DEFAULT_SETUP_PROCEDURE = 100
    TRACKER_FEEDBACK_STATE = 101
    CALIBRATION_STATE = 102
    VALIDATION_STATE = 103
    DRIFT_CORRECTION_STATE = 104

    # Pupil Measure Type Constants
    PUPIL_AREA = 70
    PUPIL_DIAMETER = 71
    PUPIL_WIDTH = 72
    PUPIL_HEIGHT = 73
    PUPIL_MAJOR_AXIS = 74
    PUPIL_MINOR_AXIS = 75
    PUPIL_RADIUS = 76
    PUPIL_DIAMETER_MM = 77
    PUPIL_WIDTH_MM = 78
    PUPIL_HEIGHT_MM = 79
    PUPIL_MAJOR_AXIS_MM = 80
    PUPIL_MINOR_AXIS_MM = 81
    PUPIL_RADIUS_MM = 82


    # Video Based Eye Tracking Algorithm Constants
    PUPIL_CR_TRACKING = 140
    PUPIL_ONLY_TRACKING = 141

    ELLIPSE_FIT = 146
    CIRCLE_FIT = 147
    CENTROID_FIT = 148

    # Eye Tracker Interface Return Code Constants
    EYETRACKER_OK = 200
    # EYETRACKER_ERROR deprecated for EYETRACKER_UNDEFINED_ERROR
    EYETRACKER_ERROR = 201
    EYETRACKER_UNDEFINED_ERROR = 201
    # FUNCTIONALITY_NOT_SUPPORTED deprecated for
    # EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED
    FUNCTIONALITY_NOT_SUPPORTED = 202
    EYETRACKER_INTERFACE_METHOD_NOT_SUPPORTED = 202
    EYETRACKER_CALIBRATION_ERROR = 203
    EYETRACKER_VALIDATION_ERROR = 204
    EYETRACKER_SETUP_ABORTED = 205
    EYETRACKER_NOT_CONNECTED = 206
    EYETRACKER_MODEL_NOT_SUPPORTED = 207
    EYETRACKER_RECEIVED_INVALID_INPUT = 208

EyeTrackerConstants.initialize()

# XInput Gamepad related

class XInputBatteryTypeConstants(Constants):
    # The device is not connected.
    BATTERY_TYPE_DISCONNECTED = 0x00
    # The device is a wired device and does not have a battery.
    BATTERY_TYPE_WIRED = 0x01
    # The device has an alkaline battery.
    BATTERY_TYPE_ALKALINE = 0x02
    # The device has a nickel metal hydride battery.
    BATTERY_TYPE_NIMH = 0x03
    # The device has an unknown battery type.
    BATTERY_TYPE_UNKNOWN = 0xFF
XInputBatteryTypeConstants.initialize()
try:
    XInputBatteryTypeConstants._keys.remove(
        XInputBatteryTypeConstants.getID('UNDEFINED'))
except Exception:  # pylint: disable=broad-except
    pass

class XInputBatteryLevelConstants(Constants):
    # BatteryLevels
    BATTERY_LEVEL_EMPTY = 0x00
    BATTERY_LEVEL_LOW = 0x01
    BATTERY_LEVEL_MEDIUM = 0x02
    BATTERY_LEVEL_FULL = 0x03

XInputBatteryLevelConstants.initialize()
try:
    XInputBatteryLevelConstants._keys.remove(
        XInputBatteryLevelConstants.getID('UNDEFINED'))
except Exception:  # pylint: disable=broad-except
    pass


class XInputCapabilitiesConstants(Constants):
    UNDEFINED = 9999999
    # Device Type
    XBOX360_GAMEPAD = 0x01
    OTHER_XINPUT_GAMEPAD = 0x0

    # subtype is defined as 0x01 in xinput.h, redining so no conflicts
    XINPUT_GAMEPAD = 0x08
    XINPUT_UNKNOWN_SUBTYPE = 0x06

XInputCapabilitiesConstants.initialize()
try:
    XInputCapabilitiesConstants._keys.remove(
        XInputCapabilitiesConstants.getID('UNDEFINED'))
except Exception:  # pylint: disable=broad-except
    pass

class XInputGamePadConstants(Constants):
    DPAD_UP = 0x0001
    DPAD_DOWN = 0x0002
    DPAD_LEFT = 0x0004
    DPAD_RIGHT = 0x0008
    START = 0x0010
    BACK = 0x0020
    LEFT_THUMB = 0x0040
    RIGHT_THUMB = 0x0080
    LEFT_SHOULDER = 0x0100
    RIGHT_SHOULDER = 0x0200
    A = 0x1000
    B = 0x2000
    X = 0x4000
    Y = 0x8000

    _batteryTypes = XInputBatteryTypeConstants()
    _batteryLevels = XInputBatteryLevelConstants()
    _capabilities = XInputCapabilitiesConstants()

XInputGamePadConstants.initialize()
try:
    XInputGamePadConstants._keys.remove(
        XInputGamePadConstants.getID('UNDEFINED'))
except Exception:  # pylint: disable=broad-except
    pass

# pylint: enable=protected-access



from . import _TestDisabledMixin, _TestBaseStandaloneRoutinesMixin
from psychopy import experiment
from psychopy.experiment.routines.eyetracker_calibrate import EyetrackerCalibrationRoutine


class TestEyetrackerCalibrationRoutine(_TestBaseStandaloneRoutinesMixin, _TestDisabledMixin):
    def setup(self):
        self.exp = experiment.Experiment()
        self.rt = EyetrackerCalibrationRoutine(exp=self.exp, name="testEyetrackerCalibrationRoutine")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Classes and functions for working with audio data.
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = [
    'AudioClip',
    'load',
    'save',
    'AUDIO_SUPPORTED_CODECS',
    'AUDIO_CHANNELS_MONO',
    'AUDIO_CHANNELS_STEREO',
    'AUDIO_CHANNEL_LEFT',
    'AUDIO_EAR_LEFT',
    'AUDIO_CHANNEL_RIGHT',
    'AUDIO_EAR_RIGHT',
    'AUDIO_CHANNEL_COUNT',
    'AUDIO_EAR_COUNT'
]

import numpy as np
import soundfile as sf
from psychopy.tools.audiotools import *
from .exceptions import *

# supported formats for loading and saving audio samples to file
AUDIO_SUPPORTED_CODECS = [s.lower() for s in sf.available_formats().keys()]

# constants for specifying the number of channels
AUDIO_CHANNELS_MONO = 1
AUDIO_CHANNELS_STEREO = 2

# constants for indexing channels
AUDIO_CHANNEL_LEFT = AUDIO_EAR_LEFT = 0
AUDIO_CHANNEL_RIGHT = AUDIO_EAR_RIGHT = 1
AUDIO_CHANNEL_COUNT = AUDIO_EAR_COUNT = 2


class AudioClip:
    """Class for storing audio clip data.

    This class is used to store and handle raw audio data, such as those
    obtained from microphone recordings or loaded from files. PsychoPy stores
    audio samples in contiguous arrays of 32-bit floating-point values ranging
    between -1 and 1.

    The `AudioClip` class provides basic audio editing capabilities too. You can
    use operators on `AudioClip` instances to combine audio clips together. For
    instance, the ``+`` operator will return a new `AudioClip` instance whose
    samples are the concatenation of the two operands::

        sndCombined = sndClip1 + sndClip2

    Note that audio clips must have the same sample rates in order to be joined
    using the addition operator. For online compatibility, use the `append()`
    method instead.

    There are also numerous static methods available to generate various tones
    (e.g., sine-, saw-, and square-waves). Audio samples can also be loaded and
    saved to files in various formats (e.g., WAV, FLAC, OGG, etc.)

    You can play `AudioClip` by directly passing instances of this object to
    the :class:`~psychopy.sound.Sound` class::

        import psychopy.core as core
        import psyhcopy.sound as sound

        myTone = AudioClip.sine(duration=5.0)  # generate a tone

        mySound = sound.Sound(myTone)
        mySound.play()
        core.wait(5.0)  # wait for sound to finish playing
        core.quit()

    Parameters
    ----------
    samples : ArrayLike
        Nx1 or Nx2 array of audio samples for mono and stereo, respectively.
        Values in the array representing the amplitude of the sound waveform
        should vary between -1 and 1. If not, they will be clipped.
    sampleRateHz : int
        Sampling rate used to obtain `samples` in Hertz (Hz). The sample rate or
        frequency is related to the quality of the audio, where higher sample
        rates usually result in better sounding audio (albeit a larger memory
        footprint and file size). The value specified should match the frequency
        the clip was recorded at. If not, the audio may sound distorted when
        played back. Usually, a sample rate of 48kHz is acceptable for most
        applications (DVD audio quality). For convenience, module level
        constants with form ``SAMPLE_RATE_*`` are provided to specify many
        common samples rates.
    userData : dict or None
        Optional user data to associated with the audio clip.

    """
    def __init__(self, samples, sampleRateHz=SAMPLE_RATE_48kHz, userData=None):
        # samples should be a 2D array where columns represent channels
        self._samples = np.atleast_2d(
            np.asarray(samples, dtype=np.float32, order='C'))
        self._samples.clip(-1, 1)  # force values to be clipped

        # set the sample rate of the clip
        self._sampleRateHz = int(sampleRateHz)

        # the duration of the audio clip
        self._duration = len(self.samples) / float(self.sampleRateHz)

        # user data
        self._userData = userData if userData is not None else {}
        assert isinstance(self._userData, dict)

    # --------------------------------------------------------------------------
    # Loading and saving
    #
    # These static methods are related to loading and saving audio clips from
    # files. The file types supported are those that `libsoundfile` supports.
    #
    # Additional codecs such as `mp3` require the pydub package which is
    # optional.
    #

    @staticmethod
    def _checkCodecSupported(codec, raiseError=False):
        """Check if the audio format string corresponds to a supported codec.
        Used internally to check if the user specified a valid codec identifier.

        Parameters
        ----------
        codec: str
            Codec identifier (e.g., 'wav', 'mp3', etc.)
        raiseError : bool
            Raise an error (``) instead of returning a value. Default is
            `False`.

        Returns
        -------
        bool
            `True` if the format is supported.

        """
        if not isinstance(codec, str):
            raise ValueError('Codec identifier must be a string.')

        hasCodec = codec.lower() in AUDIO_SUPPORTED_CODECS

        if raiseError and not hasCodec:
            fmtList = ["'{}'".format(s) for s in AUDIO_SUPPORTED_CODECS]
            raise AudioUnsupportedCodecError(
                "Unsupported audio codec specified, must be either: " +
                ", ".join(fmtList))

        return hasCodec

    @staticmethod
    def load(filename, codec=None):
        """Load audio samples from a file. Note that this is a static method!

        Parameters
        ----------
        filename : str
            File name to load.
        codec : str or None
            Codec to use. If `None`, the format will be implied from the file
            name.

        Returns
        -------
        AudioClip
            Audio clip containing samples loaded from the file.

        """
        if codec is not None:
            AudioClip._checkCodecSupported(codec, raiseError=True)

        samples, sampleRateHz = sf.read(
            filename,
            dtype='float32',
            always_2d=True,
            format=codec)

        return AudioClip(
            samples=samples,
            sampleRateHz=sampleRateHz)

    def save(self, filename, codec=None):
        """Save an audio clip to file.

        Parameters
        ----------
        filename : str
            File name to write audio clip to.
        codec : str or None
            Format to save audio clip data as. If `None`, the format will be
            implied from the extension at the end of `filename`.

        """
        if codec is not None:
            AudioClip._checkCodecSupported(codec, raiseError=True)

        sf.write(
            filename,
            data=self._samples,
            samplerate=self._sampleRateHz,
            format=codec)

    # --------------------------------------------------------------------------
    # Tone and noise generation methods
    #
    # These static methods are used to generate audio samples, such as random
    # colored noise (e.g., white) and tones (e.g., sine, square, etc.)
    #
    # All of these methods return `AudioClip` objects containing the generated
    # samples.
    #

    @staticmethod
    def whiteNoise(duration=1.0, sampleRateHz=SAMPLE_RATE_48kHz, channels=2):
        """Generate gaussian white noise.

        **New feature, use with caution.**

        Parameters
        ----------
        duration : float or int
            Length of the sound in seconds.
        sampleRateHz : int
            Samples rate of the audio for playback.
        channels : int
            Number of channels for the output.

        Returns
        -------
        AudioClip

        """
        samples = whiteNoise(duration, sampleRateHz)

        if channels > 1:
            samples = np.tile(samples, (1, channels)).astype(np.float32)

        return AudioClip(samples, sampleRateHz=sampleRateHz)

    @staticmethod
    def silence(duration=1.0, sampleRateHz=SAMPLE_RATE_48kHz, channels=2):
        """Generate audio samples for a silent period.

        This is used to create silent periods of a very specific duration
        between other audio clips.

        Parameters
        ----------
        duration : float or int
            Length of the sound in seconds.
        sampleRateHz : int
            Samples rate of the audio for playback.
        channels : int
            Number of channels for the output.

        Returns
        -------
        AudioClip

        Examples
        --------
        Generate 5 seconds of silence to enjoy::

            import psychopy.sound as sound
            silence = sound.AudioClip.silence(10.)

        Use the silence as a break between two audio clips when concatenating
        them::

            fullClip = clip1 + sound.AudioClip.silence(10.) + clip2

        """
        samples = np.zeros(
            (int(duration * sampleRateHz), channels), dtype=np.float32)

        return AudioClip(samples, sampleRateHz=sampleRateHz)

    @staticmethod
    def sine(duration=1.0, freqHz=440, gain=0.8, sampleRateHz=SAMPLE_RATE_48kHz,
             channels=2):
        """Generate audio samples for a tone with a sine waveform.

        Parameters
        ----------
        duration : float or int
            Length of the sound in seconds.
        freqHz : float or int
            Frequency of the tone in Hertz (Hz). Note that this differs from the
            `sampleRateHz`.
        gain : float
            Gain factor ranging between 0.0 and 1.0. Default is 0.8.
        sampleRateHz : int
            Samples rate of the audio for playback.
        channels : int
            Number of channels for the output.

        Returns
        -------
        AudioClip

        Examples
        --------
        Generate an audio clip of a tone 10 seconds long with a frequency of
        400Hz::

            import psychopy.sound as sound
            tone400Hz = sound.AudioClip.sine(10., 400.)

        Create a marker/cue tone and append it to pre-recorded instructions::

            import psychopy.sound as sound
            voiceInstr = sound.AudioClip.load('/path/to/instructions.wav')
            markerTone = sound.AudioClip.sine(
                1.0, 440.,  # duration and freq
                sampleRateHz=voiceInstr.sampleRateHz)  # must be the same!

            fullInstr = voiceInstr + markerTone  # create instructions with cue
            fullInstr.save('/path/to/instructions_with_tone.wav')  # save it

        """
        samples = sinetone(duration, freqHz, gain, sampleRateHz)

        if channels > 1:
            samples = np.tile(samples, (1, channels)).astype(np.float32)

        return AudioClip(samples, sampleRateHz=sampleRateHz)

    @staticmethod
    def square(duration=1.0, freqHz=440, dutyCycle=0.5, gain=0.8,
               sampleRateHz=SAMPLE_RATE_48kHz, channels=2):
        """Generate audio samples for a tone with a square waveform.

        Parameters
        ----------
        duration : float or int
            Length of the sound in seconds.
        freqHz : float or int
            Frequency of the tone in Hertz (Hz). Note that this differs from the
            `sampleRateHz`.
        dutyCycle : float
            Duty cycle between 0.0 and 1.0.
        gain : float
            Gain factor ranging between 0.0 and 1.0. Default is 0.8.
        sampleRateHz : int
            Samples rate of the audio for playback.
        channels : int
            Number of channels for the output.

        Returns
        -------
        AudioClip

        """
        samples = squaretone(duration, freqHz, dutyCycle, gain, sampleRateHz)

        if channels > 1:
            samples = np.tile(samples, (1, channels)).astype(np.float32)

        return AudioClip(samples, sampleRateHz=sampleRateHz)

    @staticmethod
    def sawtooth(duration=1.0, freqHz=440, peak=1.0, gain=0.8,
                 sampleRateHz=SAMPLE_RATE_48kHz, channels=2):
        """Generate audio samples for a tone with a sawtooth waveform.

        Parameters
        ----------
        duration : float or int
            Length of the sound in seconds.
        freqHz : float or int
            Frequency of the tone in Hertz (Hz). Note that this differs from the
            `sampleRateHz`.
        peak : float
            Location of the peak between 0.0 and 1.0. If the peak is at 0.5, the
            resulting wave will be triangular. A value of 1.0 will cause the
            peak to be located at the very end of a cycle.
        gain : float
            Gain factor ranging between 0.0 and 1.0. Default is 0.8.
        sampleRateHz : int
            Samples rate of the audio for playback.
        channels : int
            Number of channels for the output.

        Returns
        -------
        AudioClip

        """
        samples = sawtone(duration, freqHz, peak, gain, sampleRateHz)

        if channels > 1:
            samples = np.tile(samples, (1, channels)).astype(np.float32)

        return AudioClip(samples, sampleRateHz=sampleRateHz)

    # --------------------------------------------------------------------------
    # Audio editing methods
    #
    # Methods related to basic editing of audio samples (operations such as
    # splicing clips and signal gain).
    #

    def __add__(self, other):
        """Concatenate two audio clips."""
        assert other.sampleRateHz == self._sampleRateHz
        assert other.channels == self.channels

        newSamples = np.ascontiguousarray(
            np.vstack((self._samples, other.samples)),
            dtype=np.float32)

        toReturn = AudioClip(
            samples=newSamples,
            sampleRateHz=self._sampleRateHz)

        return toReturn

    def __iadd__(self, other):
        """Concatenate two audio clips inplace."""
        assert other.sampleRateHz == self._sampleRateHz
        assert other.channels == self.channels

        self._samples = np.ascontiguousarray(
            np.vstack((self._samples, other.samples)),
            dtype=np.float32)

        return self

    def append(self, clip):
        """Append samples from another sound clip to the end of this one.

        The `AudioClip` object must have the same sample rate and channels as
        this object.

        Parameters
        ----------
        clip : AudioClip
            Audio clip to append.

        Returns
        -------
        AudioClip
            This object with samples from `clip` appended.

        Examples
        --------
        Join two sound clips together::

            snd1.append(snd2)

        """
        # if either clip is empty, just replace it
        if len(self.samples) == 0:
            return clip
        if len(clip.samples) == 0:
            return self

        assert self.channels == clip.channels
        assert self._sampleRateHz == clip.sampleRateHz

        self._samples = np.ascontiguousarray(
            np.vstack((self._samples, clip.samples)),
            dtype=np.float32)

        # recompute the duration of the new clip
        self._duration = len(self.samples) / float(self.sampleRateHz)

        return self

    def copy(self):
        """Create an independent copy of this `AudioClip`.

        Returns
        -------
        AudioClip

        """
        return AudioClip(
            samples=self._samples.copy(),
            sampleRateHz=self._sampleRateHz)

    def gain(self, factor, channel=None):
        """Apply gain the audio samples.

        This will modify the internal store of samples inplace. Clipping is
        automatically applied to samples after applying gain.

        Parameters
        ----------
        factor : float or int
            Gain factor to multiply audio samples.
        channel : int or None
            Channel to apply gain to. If `None`, gain will be applied to all
            channels.

        """
        try:
            arrview = self._samples[:, :] \
                if channel is None else self._samples[:, channel]
        except IndexError:
            raise ValueError('Invalid value for `channel`.')

        # multiply and clip range
        arrview *= float(factor)
        arrview.clip(-1, 1)

    # --------------------------------------------------------------------------
    # Audio analysis methods
    #
    # Methods related to basic analysis of audio samples, nothing too advanced
    # but still useful.
    #

    def rms(self, channel=None):
        """Compute the root mean square (RMS) of the samples to determine the
        average signal level.

        Parameters
        ----------
        channel : int or None
            Channel to compute RMS (zero-indexed). If `None`, the RMS of all
            channels will be computed.

        Returns
        -------
        ndarray or float
            An array of RMS values for each channel if ``channel=None`` (even if
            there is one channel an array is returned). If `channel` *was*
            specified, a `float` will be returned indicating the RMS of that
            single channel.

        """
        if channel is not None:
            assert 0 < channel < self.channels

        arr = self._samples if channel is None else self._samples[:, channel]
        rms = np.sqrt(np.mean(np.square(arr), axis=0))

        return rms if len(rms) > 1 else rms[0]

    # --------------------------------------------------------------------------
    # Properties
    #

    @property
    def samples(self):
        """Nx1 or Nx2 array of audio samples (`~numpy.ndarray`).

        Values must range from -1 to 1. Values outside that range will be
        clipped, possibly resulting in distortion.

        """
        return self._samples

    @samples.setter
    def samples(self, value):
        self._samples = np.asarray(value, dtype=float)  # convert to array
        self._samples.clip(-1., 1.)  # do clipping to keep samples in range

        # recompute duration after updating samples
        self._duration = len(self._samples) / float(self._sampleRateHz)

    @property
    def sampleRateHz(self):
        """Sample rate of the audio clip in Hz (`int`). Should be the same
        value as the rate `samples` was captured at.

        """
        return self._sampleRateHz

    @sampleRateHz.setter
    def sampleRateHz(self, value):
        self._sampleRateHz = int(value)
        # recompute duration after updating sample rate
        self._duration = len(self._samples) / float(self._sampleRateHz)

    @property
    def duration(self):
        """The duration of the audio in seconds (`float`).

        This value is computed using the specified sampling frequency and number
        of samples.

        """
        return self._duration

    @property
    def channels(self):
        """Number of audio channels in the clip (`int`).

        If `channels` > 1, the audio clip is in stereo.

        """
        return self._samples.shape[1]

    @property
    def isStereo(self):
        """`True` if there are two channels of audio samples.

        Usually one for each ear. The first channel is usually the left ear, and
        the second the right.

        """
        return not self.isMono  # are we moving in stereo? ;)

    @property
    def isMono(self):
        """`True` if there is only one channel of audio data.

        """
        return self._samples.shape[1] == 1

    @property
    def userData(self):
        """User data associated with this clip (`dict`). Can be used for storing
        additional data related to the clip. Note that `userData` is not saved
        with audio files!

        Example
        -------
        Adding fields to `userData`. For instance, we want to associated the
        start time the clip was recorded at with it::

            myClip.userData['date_recorded'] = t_start

        We can access that field later by::

            thisRecordingStartTime = myClip.userData['date_recorded']

        """
        return self._userData

    @userData.setter
    def userData(self, value):
        assert isinstance(value, dict)
        self._userData = value

    def convertToWAV(self):
        """Get a copy of stored audio samples in WAV PCM format.

        Returns
        -------
        ndarray
            Array with the same shapes as `.samples` but in 16-bit WAV PCM
            format.

        """
        return np.asarray(
            self._samples * ((1 << 15) - 1), dtype=np.int16).tobytes()

    def asMono(self, copy=True):
        """Convert the audio clip to mono (single channel audio).

        Parameters
        ----------
        copy : bool
            If `True` an :class:`~psychopy.sound.AudioClip` containing a copy
            of the samples will be returned. If `False`, channels will be
            mixed inplace resulting a the same object being returned. User data
            is not copied.

        Returns
        -------
        :class:`~psychopy.sound.AudioClip`
            Mono version of this object.

        """
        samples = np.atleast_2d(self._samples)  # enforce 2D
        if samples.shape[1] > 1:
            samplesMixed = np.atleast_2d(
                np.sum(samples, axis=1, dtype=np.float32) / np.float32(2.)).T
        else:
            samplesMixed = samples.copy()

        if copy:
            return AudioClip(samplesMixed, self.sampleRateHz)

        self._samples = samplesMixed  # overwrite

        return self

    def transcribe(self, engine='sphinx', language='en-US', expectedWords=None,
                   config=None):
        """Convert speech in audio to text.

        This feature passes the audio clip samples to a specified text-to-speech
        engine which will attempt to transcribe any speech within. The efficacy
        of the transcription depends on the engine selected, audio quality, and
        language support. By default, Pocket Sphinx is used which provides
        decent transcription capabilities offline for English and a few other
        languages. For more robust transcription capabilities with a greater
        range of language support, online providers such as Google may be used.

        Speech-to-text conversion blocks the main application thread when used
        on Python. Don't transcribe audio during time-sensitive parts of your
        experiment! This issue is known to the developers and will be fixed in a
        later release.

        Parameters
        ----------
        engine : str
            Speech-to-text engine to use. Can be one of 'sphinx' for CMU Pocket
            Sphinx or 'google' for Google Cloud.
        language : str
            BCP-47 language code (eg., 'en-US'). Note that supported languages
            vary between transcription engines.
        expectedWords : list or tuple
            List of strings representing expected words or phrases. This will
            constrain the possible output words to the ones specified. Note not
            all engines support this feature (only Sphinx and Google Cloud do at
            this time). A warning will be logged if the engine selected does not
            support this feature. CMU PocketSphinx has an additional feature
            where the sensitivity can be specified for each expected word. You
            can indicate the sensitivity level to use by putting a ``:`` (colon)
            after each word in the list (see the Example below). Sensitivity
            levels range between 0 and 100. A higher number results in the
            engine being more conservative, resulting in a higher likelihood of
            false rejections. The default sensitivity is 80% for words/phrases
            without one specified.
        config : dict or None
            Additional configuration options for the specified engine. These
            are specified using a dictionary (ex. `config={'pfilter': 1}` will
            enable the profanity filter when using the `'google'` engine).

        Returns
        -------
        :class:`~psychopy.sound.transcribe.TranscriptionResult`
            Transcription result.

        Notes
        -----
        * Online transcription services (eg., Google) provide robust and
          accurate speech recognition capabilities with broader language support
          than offline solutions. However, these services may require a paid
          subscription to use, reliable broadband internet connections, and may
          not respect the privacy of your participants as their responses are
          being sent to a third-party. Also consider that a track of audio data
          being sent over the network can be large, users on metered connections
          may incur additional costs to run your experiment.
        * If the audio clip has multiple channels, they will be combined prior
          to being passed to the transcription service if needed.

        """
        # avoid circular import
        from psychopy.sound.transcribe import transcribe
        return transcribe(
            self,
            engine=engine,
            language=language,
            expectedWords=expectedWords,
            config=config)


def load(filename, codec=None):
    """Load an audio clip from file.

    Parameters
    ----------
    filename : str
        File name to load.
    codec : str or None
        Codec to use. If `None`, the format will be implied from the file name.

    Returns
    -------
    AudioClip
        Audio clip containing samples loaded from the file.

    """
    return AudioClip.load(filename, codec)


def save(filename, clip, codec=None):
    """Save an audio clip to file.

    Parameters
    ----------
    filename : str
        File name to write audio clip to.
    clip : AudioClip
        The clip with audio samples to write.
    codec : str or None
        Format to save audio clip data as. If `None`, the format will be
        implied from the extension at the end of `filename`.

    """
    clip.save(filename, codec)


if __name__ == "__main__":
    pass



from importlib import import_module

from ._base import BaseStandaloneRoutine, Routine
from .unknown import UnknownRoutine
from pathlib import Path


def getAllStandaloneRoutines(fetchIcons=True):
    # Safe import all modules within this folder (apart from protected ones with a _)
    for loc in Path(__file__).parent.glob("*"):
        if loc.is_dir() and not loc.name.startswith("_"):
            import_module("." + loc.name, package="psychopy.experiment.routines")
    # Get list of subclasses of BaseStandalone
    classList = BaseStandaloneRoutine.__subclasses__()
    # Remove unknown
    #if UnknownRoutine in classList:
    #    classList.remove(UnknownRoutine)
    # Get list indexed by class name with Routine removed
    classDict = {c.__name__: c for c in classList}

    return classDict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pyglet.gl import gl_info, GLint, glGetIntegerv, GL_MAX_ELEMENTS_VERTICES
from psychopy import visual, preferences
import sys, platform

print("Paths to files on the system:")
for key in ['userPrefsFile', 'appDataFile', 'demos', 'appFile']:
    print("    %s: %s" % (key, preferences.prefs.paths[key]))

print("\nSystem info:")
print(platform.platform())
if sys.platform == 'darwin':
    OSXver, junk, architecture = platform.mac_ver()
    print("macOS %s running on %s" % (OSXver, architecture))

print("\nPython info")
print(sys.executable)
print(sys.version)
import numpy
print("numpy", numpy.__version__)
import scipy
print("scipy", scipy.__version__)
import matplotlib
print("matplotlib", matplotlib.__version__)
import pyglet
print("pyglet", pyglet.version)
# pyo is a new dependency, for sound:
try:
    import pyo
    print("pyo", '%i.%i.%i' % pyo.getVersion())
except Exception:
    print('pyo [not installed]')

from psychopy import __version__
print("\nPsychoPy", __version__)

win = visual.Window([100, 100])  # some drivers want a window open first
print("have shaders:", win._haveShaders)
print("\nOpenGL info:")
# get info about the graphics card and drivers
print("vendor:", gl_info.get_vendor())
print("rendering engine:", gl_info.get_renderer())
print("OpenGL version:", gl_info.get_version())
print("(Selected) Extensions:")
extensionsOfInterest = ['GL_ARB_multitexture',
                        'GL_EXT_framebuffer_object', 'GL_ARB_fragment_program',
                        'GL_ARB_shader_objects', 'GL_ARB_vertex_shader',
                        'GL_ARB_texture_non_power_of_two', 'GL_ARB_texture_float', 'GL_STEREO']
for ext in extensionsOfInterest:
    print("\t", bool(gl_info.have_extension(ext)), ext)
# also determine nVertices that can be used in vertex arrays
maxVerts = GLint()
glGetIntegerv(GL_MAX_ELEMENTS_VERTICES, maxVerts)
print('\tmax vertices in vertex array:', maxVerts.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of dot kinematogram
"""

from psychopy import visual, event, core

win = visual.Window((600, 600), allowGUI=False, winType='pyglet')

# Initialize some stimuli
dotPatch = visual.DotStim(win, color=(1.0, 1.0, 1.0), dir=270,
    nDots=500, fieldShape='circle', fieldPos=(0.0, 0.0), fieldSize=1,
    dotLife=5,  # number of frames for each dot to be drawn
    signalDots='same',  # are signal dots 'same' on each frame? (see Scase et al)
    noiseDots='direction',  # do the noise dots follow random- 'walk', 'direction', or 'position'
    speed=0.01, coherence=0.9)

print(dotPatch)

message = visual.TextStim(win, text='Any key to quit', pos=(0, -0.5))
trialClock =core.Clock()
while not event.getKeys():
    dotPatch.draw()
    message.draw()
    win.flip()  # make the drawn things visible

    event.clearEvents('mouse')  # only really needed for pygame windows

print(win.fps())
win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
gc_cursor_demo/run.py

Demonstrates the ioHub Common EyeTracking Interface by displaying a gaze cursor
at the currently reported gaze position on an image background.

Select which tracker to use by setting the TRACKER variable below. Edit the associated
configuration dict for the eye tracker being used to modify it's settings.
"""
from psychopy import core, visual
from psychopy.data import TrialHandler, importConditions
from psychopy.iohub import launchHubServer
from psychopy.iohub.util import getCurrentDateTimeString, hideWindow, showWindow
import os

# Eye tracker to use ('mouse', 'eyelink', 'gazepoint', or 'tobii')
TRACKER = 'mouse'

eyetracker_config = dict(name='tracker')
devices_config = {}
if TRACKER == 'mouse':
    devices_config['eyetracker.hw.mouse.EyeTracker'] = eyetracker_config
    eyetracker_config['calibration'] = dict(auto_pace=True,
                                            target_duration=1.5,
                                            target_delay=1.0,
                                            screen_background_color=(0, 0, 0),
                                            type='NINE_POINTS',
                                            unit_type=None,
                                            color_type=None,
                                            target_attributes=dict(outer_diameter=0.05,
                                                                   inner_diameter=0.025,
                                                                   outer_fill_color=[-0.5, -0.5, -0.5],
                                                                   inner_fill_color=[-1, 1, -1],
                                                                   outer_line_color=[1, 1, 1],
                                                                   inner_line_color=[-1, -1, -1],
                                                                   animate=dict(enable=True,
                                                                                expansion_ratio=1.5,
                                                                                contract_only=False)
                                                                   )
                                            )
elif TRACKER == 'eyelink':
    eyetracker_config['model_name'] = 'EYELINK 1000 DESKTOP'
    eyetracker_config['simulation_mode'] = False
    eyetracker_config['runtime_settings'] = dict(sampling_rate=1000, track_eyes='RIGHT')
    eyetracker_config['calibration'] = dict(auto_pace=True,
                                            target_duration=1.5,
                                            target_delay=1.0,
                                            screen_background_color=(0, 0, 0),
                                            type='NINE_POINTS',
                                            unit_type=None,
                                            color_type=None,
                                            target_attributes=dict(outer_diameter=0.05,
                                                                   inner_diameter=0.025,
                                                                   outer_fill_color=[-0.5, -0.5, -0.5],
                                                                   inner_fill_color=[-1, 1, -1],
                                                                   outer_line_color=[1, 1, 1],
                                                                   inner_line_color=[-1, -1, -1]
                                                                   )
                                            )
    devices_config['eyetracker.hw.sr_research.eyelink.EyeTracker'] = eyetracker_config
elif TRACKER == 'gazepoint':
    eyetracker_config['device_timer'] = {'interval': 0.005}
    eyetracker_config['calibration'] = dict(use_builtin=False,
                                            target_duration=1.5,
                                            target_delay=1.0,
                                            screen_background_color=(0,0,0),
                                            type='NINE_POINTS',
                                            unit_type=None,
                                            color_type=None,
                                            target_attributes=dict(outer_diameter=0.05,
                                                                   inner_diameter=0.025,
                                                                   outer_fill_color=[-0.5, -0.5, -0.5],
                                                                   inner_fill_color=[-1, 1, -1],
                                                                   outer_line_color=[1, 1, 1],
                                                                   inner_line_color=[-1, -1, -1],
                                                                   animate=dict(enable=True,
                                                                                expansion_ratio=1.5,
                                                                                contract_only=False)
                                                                   )
                                            )
    devices_config['eyetracker.hw.gazepoint.gp3.EyeTracker'] = eyetracker_config
elif TRACKER == 'tobii':
    eyetracker_config['calibration'] = dict(auto_pace=True,
                                            target_duration=1.5,
                                            target_delay=1.0,
                                            screen_background_color=(0, 0, 0),
                                            type='NINE_POINTS',
                                            unit_type=None,
                                            color_type=None,
                                            target_attributes=dict(outer_diameter=0.05,
                                                                   inner_diameter=0.025,
                                                                   outer_fill_color=[-0.5, -0.5, -0.5],
                                                                   inner_fill_color=[-1, 1, -1],
                                                                   outer_line_color=[1, 1, 1],
                                                                   inner_line_color=[-1, -1, -1],
                                                                   animate=dict(enable=True,
                                                                                expansion_ratio=1.5,
                                                                                contract_only=False)
                                                                   )
                                            )
    devices_config['eyetracker.hw.tobii.EyeTracker'] = eyetracker_config
else:
    print("{} is not a valid TRACKER name; please use 'mouse', 'eyelink', 'gazepoint', or 'tobii'.".format(TRACKER))
    core.quit()

if __name__ == "__main__":
    window = visual.Window((1920, 1080),
                        units='height',
                        fullscr=True,
                        allowGUI=False,
                        colorSpace='rgb',
                        color=[0, 0, 0]
                        )

    window.setMouseVisible(False)

    # Create a dict of image stim for trials and a gaze blob to show the
    # reported gaze position with.
    #
    image_cache = dict()
    image_names = ['canal.jpg', 'fall.jpg', 'party.jpg', 'swimming.jpg', 'lake.jpg']
    for iname in image_names:
        image_cache[iname] = visual.ImageStim(window, image=os.path.join('./images/', iname), name=iname)

    # Create a circle to use for the Gaze Cursor. Current units assume pix.
    #
    gaze_dot = visual.GratingStim(window, tex=None, mask="gauss", pos=(0, 0),
                                  size=(0.1, 0.1), color='green')

    # Create a Text Stim for use on /instruction/ type screens.
    # Current units assume pix.
    instructions_text_stim = visual.TextStim(window, text='', pos=[0, 0], units='pix', height=24, color=[-1, -1, -1],
                                             wrapWidth=window.size[0]*.9)


    exp_conditions = importConditions('trial_conditions.xlsx')
    trials = TrialHandler(exp_conditions, 1)

    io_hub = launchHubServer(window=window, experiment_code='gc_cursor', **devices_config)
    # Inform the ioDataStore that the experiment is using a TrialHandler. The ioDataStore will create a table
    # which can be used to record the actual trial variable values (DV or IV) in the order run / collected.
    #
    io_hub.createTrialHandlerRecordTable(trials)

    # Let's make some short-cuts to the devices we will be using in this demo.
    tracker = None
    try:
        tracker = io_hub.devices.tracker
    except Exception:
        print(" No eye tracker config found in iohub_config.yaml")
        io_hub.quit()
        core.quit()

    display = io_hub.devices.display
    kb = io_hub.devices.keyboard

    # Minimize the PsychoPy window if needed
    hideWindow(window)
    # Display calibration gfx window and run calibration.
    result = tracker.runSetupProcedure()
    print("Calibration returned: ", result)
    # Maximize the PsychoPy window if needed
    showWindow(window)

    flip_time = window.flip()
    io_hub.sendMessageEvent(text="EXPERIMENT_START", sec_time=flip_time)

    # Send some information to the ioDataStore as experiment messages,
    # including the experiment and session id's, the calculated pixels per
    # degree, display resolution, etc.
    #
    io_hub.sendMessageEvent(text="IO_HUB EXPERIMENT_INFO START")
    io_hub.sendMessageEvent(text="ioHub Experiment started {0}".format(getCurrentDateTimeString()))
    io_hub.sendMessageEvent(text="Experiment ID: {0}, Session ID: {1}".format(io_hub.experimentID,
                                                                              io_hub.experimentSessionID))
    io_hub.sendMessageEvent(text="Stimulus Screen ID: {0}, "
                            "Size (pixels): {1}, CoordType: {2}".format(display.getIndex(),
                                                                        display.getPixelResolution(),
                                                                        display.getCoordinateType()))
    io_hub.sendMessageEvent(text="Calculated Pixels Per Degree: {0} x, {1} y".format(*display.getPixelsPerDegree()))
    io_hub.sendMessageEvent(text="IO_HUB EXPERIMENT_INFO END")

    io_hub.clearEvents('all')

    # For each trial in the set of trials within the current block.
    #
    t = 0
    for trial in trials:
        # Update the instruction screen text to indicate
        # a trial is about to start.
        #
        instuction_text = "Press Space Key To Start Trial %d" % t
        instructions_text_stim.setText(instuction_text)
        instructions_text_stim.draw()
        window.flip()

        # Wait until a space key press event occurs after the
        # start trial instuctions have been displayed.
        #
        io_hub.clearEvents('all')
        kb.waitForPresses(keys=[' ', ])

        # Space Key has been pressed, start the trial.
        # Set the current session and trial id values to be saved
        # in the ioDataStore for the upcoming trial.
        #

        trial['session_id'] = io_hub.getSessionID()
        trial['trial_id'] = t+1

        # Start Recording Eye Data
        #
        tracker.setRecordingState(True)

        # Get the image stim for this trial.
        #
        imageStim = image_cache[trial['IMAGE_NAME']]
        imageStim.draw()
        flip_time = window.flip()
        # Clear all the events received prior to the trial start.
        #
        io_hub.clearEvents('all')
        # Send a msg to the ioHub indicating that the trial started,
        # and the time of the first retrace displaying the trial stim.
        #
        io_hub.sendMessageEvent(text="TRIAL_START", sec_time=flip_time)
        # Set the value of the trial start variable for this trial
        #
        trial['TRIAL_START'] = flip_time

        # Loop until we get a keyboard event
        #
        run_trial = True
        while run_trial is True:
            # Get the latest gaze position in display coord space..
            #
            gpos = tracker.getPosition()
            if type(gpos) in [tuple, list]:
                # If we have a gaze position from the tracker,
                # redraw the background image and then the
                # gaze_cursor at the current eye position.
                #
                gaze_dot.setPos([gpos[0], gpos[1]])
                imageStim.draw()
                gaze_dot.draw()
            else:
                # Otherwise just draw the background image.
                # This will remove the gaze cursor from the screen
                # when the eye tracker is not successfully
                # tracking eye position.
                #
                imageStim.draw()

            # Flip video buffers, displaying the stim we just
            # updated.
            #
            flip_time = window.flip()

            # Send an experiment message to the ioDataStore
            # indicating the time the image was drawn and
            # current position of gaze spot.
            #
            if type(gpos) in [tuple, list]:
                io_hub.sendMessageEvent("IMAGE_UPDATE %s %.3f %.3f" % (trial['IMAGE_NAME'], gpos[0], gpos[1]),
                                        sec_time=flip_time)
            else:
                io_hub.sendMessageEvent("IMAGE_UPDATE %s [NO GAZE]" % (trial['IMAGE_NAME']),
                                        sec_time=flip_time)

            # Check any new keyboard press events by a space key.
            # If one is found, set the trial end variable and break.
            # from the loop
            if kb.getPresses(keys=[' ', ]):
                run_trial = False
                break

        # The trial has ended, so update the trial end time condition value,
        # and send a message to the ioDataStore with the trial end time.
        #
        flip_time = window.flip()
        trial['TRIAL_END'] = flip_time
        io_hub.sendMessageEvent(text="TRIAL_END", sec_time=flip_time)

        # Stop recording eye data.
        # In this example, we have no use for any eye data
        # between trials, so why save it.
        #
        tracker.setRecordingState(False)

        # Save the experiment condition variable values for this
        # trial to the ioDataStore.
        #
        io_hub.addTrialHandlerRecord(trial)

        # Clear all event buffers
        #
        io_hub.clearEvents('all')
        t += 1

    # All trials have been run, so end the experiment.
    #

    flip_time = window.flip()
    io_hub.sendMessageEvent(text='EXPERIMENT_COMPLETE', sec_time=flip_time)

    # Disconnect the eye tracking device.
    #
    tracker.setConnectionState(False)

    # The experiment is done, all trials have been run.
    # Clear the screen and show an 'experiment  done' message using the
    # instructionScreen text.
    #
    instuction_text = "Press Any Key to Exit Demo"
    instructions_text_stim.setText(instuction_text)
    instructions_text_stim.draw()
    flip_time = window.flip()
    io_hub.sendMessageEvent(text="SHOW_DONE_TEXT", sec_time=flip_time)
    io_hub.clearEvents('all')
    # wait until any key is pressed
    kb.waitForPresses()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2015 Jonathan Peirce
# Distributed under the terms of the GNU General Public License (GPL).

from pathlib import Path
from psychopy.experiment.components import BaseVisualComponent, Param, \
    getInitVals, _translate
from psychopy.visual import slider
from psychopy.experiment import py2js
from psychopy import logging
from psychopy.data import utils
from psychopy.localization import _localized as __localized
_localized = __localized.copy()
import copy

__author__ = 'Jon Peirce'

# only use _localized values for label values, nothing functional:
_localized.update({'categoryChoices': _translate('Category choices'),
                   'labels': _translate('Labels'),
                   'ticks': _translate('Ticks'),
                   'forceEndRoutine': _translate('Force end of Routine'),
                   'storeHistory': _translate('Store history'),
                   'storeRating': _translate('Store rating'),
                   'storeRatingTime': _translate('Store rating time'),
                   'readOnly': _translate('readOnly')})

knownStyles = slider.Slider.knownStyles
legacyStyles = slider.Slider.legacyStyles
knownStyleTweaks = slider.Slider.knownStyleTweaks
legacyStyleTweaks = slider.Slider.legacyStyleTweaks


# ticks = (1, 2, 3, 4, 5),
# labels = None,
# pos = None,
# size = None,
# units = None,
# flip = False,
# style = 'rating',
# granularity = 0,
# textSize = 1.0,
# readOnly = False,
# color = 'LightGray',
# textFont = 'Helvetica Bold',

class SliderComponent(BaseVisualComponent):
    """A class for presenting a rating scale as a builder component
    """

    categories = ['Responses']
    targets = ['PsychoPy', 'PsychoJS']
    iconFile = Path(__file__).parent / 'slider.png'
    tooltip = _translate('Slider: A simple, flexible object for getting ratings')

    def __init__(self, exp, parentName,
                 name='slider',
                 labels='',
                 ticks="(1, 2, 3, 4, 5)",
                 initVal="",
                 size='(1.0, 0.1)',
                 pos='(0, -0.4)',
                 flip=False,
                 style='rating', styleTweaks=[],
                 granularity=0,
                 color="LightGray",
                 fillColor='Red',
                 borderColor='White',
                 font="Open Sans",
                 letterHeight=0.05,
                 startType='time (s)', startVal='0.0',
                 stopType='condition', stopVal='',
                 startEstim='', durationEstim='',
                 forceEndRoutine=True,
                 storeRating=True, storeRatingTime=True, storeHistory=False, readOnly=False):
        super(SliderComponent, self).__init__(
                exp, parentName, name,
                pos=pos, size=size,
                color=color, fillColor=fillColor, borderColor=borderColor,
                startType=startType, startVal=startVal,
                stopType=stopType, stopVal=stopVal,
                startEstim=startEstim, durationEstim=durationEstim)
        self.type = 'Slider'
        self.url = "https://www.psychopy.org/builder/components/slider.html"
        self.exp.requirePsychopyLibs(['visual', 'event'])

        # params
        self.order += ['forceEndRoutine',  # Basic tab
                       'contrast', 'styles', 'styleTweaks', # Appearance tab
                       'font',  # Formatting tab
                       'flip',  # Layout tab
                       'ticks', 'labels',  'granularity', 'readOnly',  # Data tab
                      ]
        self.order.insert(self.order.index("colorSpace"), "style")
        self.order.insert(self.order.index("units"), "Item Padding")

        # normal params:
        # = the usual as inherited from BaseVisual plus:
        self.params['ticks'] = Param(
                ticks, valType='list', inputType="single", allowedTypes=[], categ='Basic',
                updates='constant',
                hint=_translate("Tick positions (numerical) on the scale, "
                                "separated by commas"),
                label=_localized['ticks'])
        self.params['labels'] = Param(
                labels, valType='list', inputType="single", allowedTypes=[], categ='Basic',
                updates='constant',
                hint=_translate("Labels for the tick marks on the scale, "
                                "separated by commas"),
                label=_localized['labels'])
        self.params['initVal'] = Param(
            initVal, valType='code', inputType="single", categ='Basic',
            hint=_translate("Value of the slider befre any response, leave blank to hide the marker until clicked on"),
            label=_translate("Starting Value")
        )
        self.params['granularity'] = Param(
                granularity, valType='num', inputType="single", allowedTypes=[], categ='Basic',
                updates='constant',
                hint=_translate("Specifies the minimum step size "
                                "(0 for a continuous scale, 1 for integer "
                                "rating scale)"),
                label=_translate('Granularity'))
        self.params['forceEndRoutine'] = Param(
                forceEndRoutine, valType='bool', inputType="bool", allowedTypes=[], categ='Basic',
                updates='constant', allowedUpdates=[],
                hint=_translate("Should setting a rating (releasing the mouse) "
                                "cause the end of the routine (e.g. trial)?"),
                label=_localized['forceEndRoutine'])
        self.params['readOnly'] = Param(
            readOnly, valType='bool', allowedTypes=[], categ='Data',
            updates='constant', allowedUpdates=[],
            hint=_translate("Should participant be able to change the rating on the Slider?"),
            label=_localized['readOnly'])

        # advanced params:
        self.params['flip'] = Param(
                flip, valType='bool', inputType="bool", categ='Layout',
                updates='constant', allowedUpdates=[],
                hint=_translate(
                        "By default the labels will be on the bottom or "
                        "left of the scale, but this can be flipped to the "
                        "other side."),
                label=_translate('Flip'))

        # Color changes
        self.params['color'].label = _translate("Label Color")
        self.params['color'].hint = _translate("Color of all labels on this slider (might be overridden by the style setting)")
        self.params['fillColor'].label = _translate("Marker Color")
        self.params['fillColor'].hint = _translate("Color of the marker on this slider (might be overridden by the style setting)")
        self.params['borderColor'].label = _translate("Line Color")
        self.params['borderColor'].hint = _translate("Color of all lines on this slider (might be overridden by the style setting)")

        self.params['font'] = Param(
                font, valType='str', inputType="single", categ='Formatting',
                updates='constant',
                allowedUpdates=['constant', 'set every repeat'],
                hint=_translate(
                        "Font for the labels"),
                label=_translate('Font'))

        self.params['letterHeight'] = Param(
                letterHeight, valType='num', inputType="single", categ='Formatting',
                updates='constant',
                allowedUpdates=['constant', 'set every repeat'],
                hint=_translate(
                        "Letter height for text in labels"),
                label=_translate('Letter height'))

        self.params['styles'] = Param(
                style, valType='str', inputType="choice", categ='Appearance',
                updates='constant', allowedVals=knownStyles,
                hint=_translate(
                        "Discrete styles to control the overall appearance of the slider."),
                label=_translate('Styles'))

        self.params['styleTweaks'] = Param(
                styleTweaks, valType='list', inputType="multiChoice", categ='Appearance',
                updates='constant', allowedVals=knownStyleTweaks,
                hint=_translate(
                        "Tweaks to change the appearance of the slider beyond its style."),
                label=_translate('Style Tweaks'))

        # data params
        self.params['storeRating'] = Param(
                storeRating, valType='bool', inputType="bool", allowedTypes=[], categ='Data',
                updates='constant', allowedUpdates=[],
                hint=_translate("store the rating"),
                label=_localized['storeRating'])
        self.params['storeRatingTime'] = Param(
                storeRatingTime, valType='bool', inputType="bool", allowedTypes=[], categ='Data',
                updates='constant', allowedUpdates=[],
                hint=_translate("Store the time taken to make the choice (in "
                                "seconds)"),
                label=_localized['storeRatingTime'])
        self.params['storeHistory'] = Param(
                storeHistory, valType='bool', inputType="bool", allowedTypes=[], categ='Data',
                updates='constant', allowedUpdates=[],
                hint=_translate("store the history of (selection, time)"),
                label=_localized['storeHistory'])

    def writeInitCode(self, buff):

        inits = getInitVals(self.params)
        # check units
        if inits['units'].val == 'from exp settings':
            inits['units'].val = None

        inits['depth'] = -self.getPosInRoutine()

        # Use None as a start value if none set
        inits['initVal'] = inits['initVal'] or None

        # build up an initialization string for Slider():
        initStr = ("{name} = visual.Slider(win=win, name='{name}',\n"
                   "    startValue={initVal}, size={size}, pos={pos}, units={units},\n"
                   "    labels={labels}, ticks={ticks}, granularity={granularity},\n"
                   "    style={styles}, styleTweaks={styleTweaks}, opacity={opacity},\n"
                   "    labelColor={color}, markerColor={fillColor}, lineColor={borderColor}, colorSpace={colorSpace},\n"
                   "    font={font}, labelHeight={letterHeight},\n"
                   "    flip={flip}, ori={ori}, depth={depth}, readOnly={readOnly})\n"
                   .format(**inits))
        buff.writeIndented(initStr)

    def writeInitCodeJS(self, buff):
        inits = getInitVals(self.params)
        for param in inits:
            if inits[param].val in ['', None, 'None', 'none']:
                inits[param].val = 'undefined'

        # Check for unsupported units
        if inits['units'].val == 'from exp settings':
            inits['units'] = copy.copy(self.exp.settings.params['Units'])
        if inits['units'].val in ['cm', 'deg', 'degFlatPos', 'degFlat']:
            msg = ("'{units}' units for your '{name}' Slider are not currently supported for PsychoJS: "
                  "switching units to 'height'. Note, this will affect the size and positioning of '{name}'.")
            logging.warning(msg.format(units=inits['units'].val, name=inits['name'].val))
            inits['units'].val = "height"

        boolConverter = {False: 'false', True: 'true'}
        sliderStyles = {'slider': 'SLIDER',
                        'scrollbar': 'SLIDER',
                        '()': 'RATING',
                        'rating': 'RATING',
                        'radio': 'RADIO',
                        'labels45': 'LABELS_45',
                        'whiteOnBlack': 'WHITE_ON_BLACK',
                        'triangleMarker': 'TRIANGLE_MARKER'}

        # If no style given, set default 'rating' as list
        if len(inits['styles'].val) == 0:
            inits['styles'].val = 'rating'

        # reformat styles for JS
        # concatenate styles and tweaks
        tweaksList = utils.listFromString(self.params['styleTweaks'].val)
        if type(inits['styles'].val) == list:  # from an experiment <2021.1
            stylesList = inits['styles'].val + tweaksList
        else:
            stylesList = [inits['styles'].val] + tweaksList
        stylesListJS = [sliderStyles[this] for this in stylesList]
        # if not isinstance(inits['styleTweaks'].val, (tuple, list)):
        #     inits['styleTweaks'].val = [inits['styleTweaks'].val]
        # inits['styleTweaks'].val = ', '.join(["visual.Slider.StyleTweaks.{}".format(adj)
        #                                       for adj in inits['styleTweaks'].val])

        # convert that to string and JS-ify
        inits['styles'].val = py2js.expression2js(str(stylesListJS))
        inits['styles'].valType = 'code'

        inits['depth'] = -self.getPosInRoutine()

        # build up an initialization string for Slider():
        initStr = ("{name} = new visual.Slider({{\n"
                   "  win: psychoJS.window, name: '{name}',\n"
                   "  size: {size}, pos: {pos}, units: {units},\n"
                   "  labels: {labels}, fontSize: {letterHeight}, ticks: {ticks},\n"
                   "  granularity: {granularity}, style: {styles},\n"
                   "  color: new util.Color({color}), markerColor: new util.Color({fillColor}), lineColor: new util.Color({borderColor}), \n"
                   "  fontFamily: {font}, bold: true, italic: false, depth: {depth}, \n"
                   ).format(**inits)
        initStr += ("  flip: {flip},\n"
                    "}});\n\n").format(flip=boolConverter[inits['flip'].val])
        buff.writeIndentedLines(initStr)

    def writeRoutineStartCode(self, buff):
        buff.writeIndented("%(name)s.reset()\n" % (self.params))

    def writeRoutineStartCodeJS(self, buff):
        buff.writeIndented("%(name)s.reset()\n" % (self.params))

    def writeFrameCode(self, buff):
        super(SliderComponent, self).writeFrameCode(buff)  # Write basevisual frame code
        forceEnd = self.params['forceEndRoutine'].val
        if forceEnd:
            code = ("\n# Check %(name)s for response to end routine\n"
                    "if %(name)s.getRating() is not None and %(name)s.status == STARTED:\n"
                    "    continueRoutine = False")
            buff.writeIndentedLines(code % (self.params))

    def writeFrameCodeJS(self, buff):
        super(SliderComponent, self).writeFrameCodeJS(buff)  # Write basevisual frame code
        forceEnd = self.params['forceEndRoutine'].val
        if forceEnd:
            code = ("\n// Check %(name)s for response to end routine\n"
                    "if (%(name)s.getRating() !== undefined && %(name)s.status === PsychoJS.Status.STARTED) {\n"
                    "  continueRoutine = false; }\n")
            buff.writeIndentedLines(code % (self.params))

    def writeRoutineEndCode(self, buff):
        name = self.params['name']
        if len(self.exp.flow._loopList):
            currLoop = self.exp.flow._loopList[-1]  # last (outer-most) loop
        else:
            currLoop = self.exp._expHandler

        # write the actual code
        storeTime = self.params['storeRatingTime'].val
        if self.params['storeRating'].val or storeTime:
            if currLoop.type in ['StairHandler', 'QuestHandler']:
                msg = ("# NB PsychoPy doesn't handle a 'correct answer' "
                       "for Slider events so doesn't know what to "
                       "tell a StairHandler (or QuestHandler)\n")
                buff.writeIndented(msg)
            elif currLoop.type in ['TrialHandler', 'ExperimentHandler']:
                loopName = currLoop.params['name']
            else:
                loopName = 'thisExp'

            if self.params['storeRating'].val == True:
                code = "%s.addData('%s.response', %s.getRating())\n"
                buff.writeIndented(code % (loopName, name, name))
            if self.params['storeRatingTime'].val == True:
                code = "%s.addData('%s.rt', %s.getRT())\n"
                buff.writeIndented(code % (loopName, name, name))
            if self.params['storeHistory'].val == True:
                code = "%s.addData('%s.history', %s.getHistory())\n"
                buff.writeIndented(code % (loopName, name, name))

            # get parent to write code too (e.g. store onset/offset times)
            super().writeRoutineEndCode(buff)

    def writeRoutineEndCodeJS(self, buff):
        name = self.params['name']
        if len(self.exp.flow._loopList):
            currLoop = self.exp.flow._loopList[-1]  # last (outer-most) loop
        else:
            currLoop = self.exp._expHandler

        # write the actual code
        storeTime = self.params['storeRatingTime'].val
        if self.params['storeRating'].val or storeTime:
            if currLoop.type in ['StairHandler', 'QuestHandler']:
                msg = ("/* NB PsychoPy doesn't handle a 'correct answer' "
                       "for Slider events so doesn't know what to "
                       "tell a StairHandler (or QuestHandler)*/\n")
                buff.writeIndented(msg)

            if self.params['storeRating'].val == True:
                code = "psychoJS.experiment.addData('%s.response', %s.getRating());\n"
                buff.writeIndented(code % (name, name))
            if self.params['storeRatingTime'].val == True:
                code = "psychoJS.experiment.addData('%s.rt', %s.getRT());\n"
                buff.writeIndented(code % (name, name))
            if self.params['storeHistory'].val == True:
                code = "psychoJS.experiment.addData('%s.history', %s.getHistory());\n"
                buff.writeIndented(code % (name, name))



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# Initial file based on pyglet.libs.win32
# Tablet related class definitions
import pyglet
from pyglet.event import EventDispatcher
import ctypes
from ...errors import print2err
from ...constants import EventConstants

lib = ctypes.windll.wintab32

LONG = ctypes.c_long
BOOL = ctypes.c_int
UINT = ctypes.c_uint
WORD = ctypes.c_uint16
DWORD = ctypes.c_uint32
WCHAR = ctypes.c_wchar
FIX32 = DWORD
WTPKT = DWORD

LCNAMELEN = 40


def print_struct(s, pretxt=''):
    print2err('{}{}:'.format(pretxt, s))
    for field_name, field_type in s._fields_:
        print2err('\t{}:\t{}'.format(field_name, getattr(s, field_name)))
    print2err('<<<<<<<<')


def struct2dict(s):
    csdict = {}
    for field_name, field_type in s._fields_:
        ufield_name = field_name
        if field_name.startswith('ax'):
            ufield_name = field_name[2:].lower()
        csdict[ufield_name] = getattr(s, field_name)
    return csdict


class AXIS(ctypes.Structure):
    _fields_ = (
        ('axMin', LONG),
        ('axMax', LONG),
        ('axUnits', UINT),
        ('axResolution', FIX32)
    )

    def get_scale(self):
        return 1 / float(self.axMax - self.axMin)

    def get_bias(self):
        return -self.axMin


class ORIENTATION(ctypes.Structure):
    _fields_ = (
        ('orAzimuth', ctypes.c_int),
        ('orAltitude', ctypes.c_int),
        ('orTwist', ctypes.c_int)
    )


class ROTATION(ctypes.Structure):
    _fields_ = (
        ('roPitch', ctypes.c_int),
        ('roRoll', ctypes.c_int),
        ('roYaw', ctypes.c_int),
    )


class LOGCONTEXT(ctypes.Structure):
    _fields_ = (
        ('lcName', WCHAR * LCNAMELEN),
        ('lcOptions', UINT),
        ('lcStatus', UINT),
        ('lcLocks', UINT),
        ('lcMsgBase', UINT),
        ('lcDevice', UINT),
        ('lcPktRate', UINT),
        ('lcPktData', WTPKT),
        ('lcPktMode', WTPKT),
        ('lcMoveMask', WTPKT),
        ('lcBtnDnMask', DWORD),
        ('lcBtnUpMask', DWORD),
        ('lcInOrgX', LONG),
        ('lcInOrgY', LONG),
        ('lcInOrgZ', LONG),
        ('lcInExtX', LONG),
        ('lcInExtY', LONG),
        ('lcInExtZ', LONG),
        ('lcOutOrgX', LONG),
        ('lcOutOrgY', LONG),
        ('lcOutOrgZ', LONG),
        ('lcOutExtX', LONG),
        ('lcOutExtY', LONG),
        ('lcOutExtZ', LONG),
        ('lcSensX', FIX32),
        ('lcSensY', FIX32),
        ('lcSensZ', FIX32),
        ('lcSysMode', BOOL),
        ('lcSysOrgX', ctypes.c_int),
        ('lcSysOrgY', ctypes.c_int),
        ('lcSysExtX', ctypes.c_int),
        ('lcSysExtY', ctypes.c_int),
        ('lcSysSensX', FIX32),
        ('lcSysSensY', FIX32),
    )


PK_CONTEXT = 0x0001  # reporting context
PK_STATUS = 0x0002  # status bits
PK_TIME = 0x0004  # time stamp
PK_CHANGED = 0x0008  # change bit vector
PK_SERIAL_NUMBER = 0x0010  # packet serial number
PK_CURSOR = 0x0020  # reporting cursor
PK_BUTTONS = 0x0040  # button information
PK_X = 0x0080  # x axis
PK_Y = 0x0100  # y axis
PK_Z = 0x0200  # z axis
PK_NORMAL_PRESSURE = 0x0400  # normal or tip pressure
PK_TANGENT_PRESSURE = 0x0800  # tangential or barrel pressure
PK_ORIENTATION = 0x1000  # orientation info: tilts
PK_ROTATION = 0x2000  # rotation info; 1.1

DEFAULT_PACKET_DATA_FIELDS = (
    PK_STATUS | PK_TIME | PK_CHANGED |
    PK_SERIAL_NUMBER | PK_CURSOR | PK_BUTTONS |
    PK_X | PK_Y | PK_Z |
    PK_NORMAL_PRESSURE |
    PK_ORIENTATION)


class PACKET(ctypes.Structure):
    _fields_ = (
        ('pkStatus', UINT),
        ('pkTime', LONG),
        ('pkChanged', WTPKT),
        ('pkSerialNumber', UINT),
        ('pkCursor', UINT),
        ('pkButtons', DWORD),
        ('pkX', LONG),
        ('pkY', LONG),
        ('pkZ', LONG),
        ('pkNormalPressure', UINT),
        ('pkOrientation', ORIENTATION)
    )

TU_NONE = 0
TU_INCHES = 1
TU_CENTIMETERS = 2
TU_CIRCLE = 3

# messages
WT_DEFBASE = 0x7ff0
WT_MAXOFFSET = 0xf
WT_PACKET = 0  # remember to add base
WT_CTXOPEN = 1
WT_CTXCLOSE = 2
WT_CTXUPDATE = 3
WT_CTXOVERLAP = 4
WT_PROXIMITY = 5
WT_INFOCHANGE = 6
WT_CSRCHANGE = 7

# system button assignment values
SBN_NONE = 0x00
SBN_LCLICK = 0x01
SBN_LDBLCLICK = 0x02
SBN_LDRAG = 0x03
SBN_RCLICK = 0x04
SBN_RDBLCLICK = 0x05
SBN_RDRAG = 0x06
SBN_MCLICK = 0x07
SBN_MDBLCLICK = 0x08
SBN_MDRAG = 0x09

# for Pen Windows
SBN_PTCLICK = 0x10
SBN_PTDBLCLICK = 0x20
SBN_PTDRAG = 0x30
SBN_PNCLICK = 0x40
SBN_PNDBLCLICK = 0x50
SBN_PNDRAG = 0x60
SBN_P1CLICK = 0x70
SBN_P1DBLCLICK = 0x80
SBN_P1DRAG = 0x90
SBN_P2CLICK = 0xA0
SBN_P2DBLCLICK = 0xB0
SBN_P2DRAG = 0xC0
SBN_P3CLICK = 0xD0
SBN_P3DBLCLICK = 0xE0
SBN_P3DRAG = 0xF0

HWC_INTEGRATED = 0x0001
HWC_TOUCH = 0x0002
HWC_HARDPROX = 0x0004
HWC_PHYSID_CURSORS = 0x0008  # 1.1

CRC_MULTIMODE = 0x0001  # 1.1
CRC_AGGREGATE = 0x0002  # 1.1
CRC_INVERT = 0x0004  # 1.1

WTI_INTERFACE = 1
IFC_WINTABID = 1
IFC_SPECVERSION = 2
IFC_IMPLVERSION = 3
IFC_NDEVICES = 4
IFC_NCURSORS = 5
IFC_NCONTEXTS = 6
IFC_CTXOPTIONS = 7
IFC_CTXSAVESIZE = 8
IFC_NEXTENSIONS = 9
IFC_NMANAGERS = 10
IFC_MAX = 10

WTI_STATUS = 2
STA_CONTEXTS = 1
STA_SYSCTXS = 2
STA_PKTRATE = 3
STA_PKTDATA = 4
STA_MANAGERS = 5
STA_SYSTEM = 6
STA_BUTTONUSE = 7
STA_SYSBTNUSE = 8
STA_MAX = 8

WTI_DEFCONTEXT = 3
WTI_DEFSYSCTX = 4
WTI_DDCTXS = 400  # 1.1
WTI_DSCTXS = 500  # 1.1
CTX_NAME = 1
CTX_OPTIONS = 2
CTX_STATUS = 3
CTX_LOCKS = 4
CTX_MSGBASE = 5
CTX_DEVICE = 6
CTX_PKTRATE = 7
CTX_PKTDATA = 8
CTX_PKTMODE = 9
CTX_MOVEMASK = 10
CTX_BTNDNMASK = 11
CTX_BTNUPMASK = 12
CTX_INORGX = 13
CTX_INORGY = 14
CTX_INORGZ = 15
CTX_INEXTX = 16
CTX_INEXTY = 17
CTX_INEXTZ = 18
CTX_OUTORGX = 19
CTX_OUTORGY = 20
CTX_OUTORGZ = 21
CTX_OUTEXTX = 22
CTX_OUTEXTY = 23
CTX_OUTEXTZ = 24
CTX_SENSX = 25
CTX_SENSY = 26
CTX_SENSZ = 27
CTX_SYSMODE = 28
CTX_SYSORGX = 29
CTX_SYSORGY = 30
CTX_SYSEXTX = 31
CTX_SYSEXTY = 32
CTX_SYSSENSX = 33
CTX_SYSSENSY = 34
CTX_MAX = 34

WTI_DEVICES = 100
DVC_NAME = 1
DVC_HARDWARE = 2
DVC_NCSRTYPES = 3
DVC_FIRSTCSR = 4
DVC_PKTRATE = 5
DVC_PKTDATA = 6
DVC_PKTMODE = 7
DVC_CSRDATA = 8
DVC_XMARGIN = 9
DVC_YMARGIN = 10
DVC_ZMARGIN = 11
DVC_X = 12
DVC_Y = 13
DVC_Z = 14
DVC_NPRESSURE = 15
DVC_TPRESSURE = 16
DVC_ORIENTATION = 17
DVC_ROTATION = 18  # 1.1
DVC_PNPID = 19  # 1.1
DVC_MAX = 19

WTI_CURSORS = 200
CSR_NAME = 1
CSR_ACTIVE = 2
CSR_PKTDATA = 3
CSR_BUTTONS = 4
CSR_BUTTONBITS = 5
CSR_BTNNAMES = 6
CSR_BUTTONMAP = 7
CSR_SYSBTNMAP = 8
CSR_NPBUTTON = 9
CSR_NPBTNMARKS = 10
CSR_NPRESPONSE = 11
CSR_TPBUTTON = 12
CSR_TPBTNMARKS = 13
CSR_TPRESPONSE = 14
CSR_PHYSID = 15  # 1.1
CSR_MODE = 16  # 1.1
CSR_MINPKTDATA = 17  # 1.1
CSR_MINBUTTONS = 18  # 1.1
CSR_CAPABILITIES = 19  # 1.1
CSR_TYPE = 20  # 1.2
CSR_MAX = 20

WTI_EXTENSIONS = 300
EXT_NAME = 1
EXT_TAG = 2
EXT_MASK = 3
EXT_SIZE = 4
EXT_AXES = 5
EXT_DEFAULT = 6
EXT_DEFCONTEXT = 7
EXT_DEFSYSCTX = 8
EXT_CURSORS = 9
EXT_MAX = 109  # Allow 100 cursors
CXO_SYSTEM = 0x0001
CXO_PEN = 0x0002
CXO_MESSAGES = 0x0004
CXO_MARGIN = 0x8000
CXO_MGNINSIDE = 0x4000
CXO_CSRMESSAGES = 0x0008  # 1.1

# context status values
CXS_DISABLED = 0x0001
CXS_OBSCURED = 0x0002
CXS_ONTOP = 0x0004

# context lock values
CXL_INSIZE = 0x0001
CXL_INASPECT = 0x0002
CXL_SENSITIVITY = 0x0004
CXL_MARGIN = 0x0008
CXL_SYSOUT = 0x0010
# packet status values
TPS_PROXIMITY = 0x0001
TPS_QUEUE_ERR = 0x0002
TPS_MARGIN = 0x0004
TPS_GRAB = 0x0008
TPS_INVERT = 0x0010  # 1.1

TBN_NONE = 0
TBN_UP = 1
TBN_DOWN = 2
PKEXT_ABSOLUTE = 1
PKEXT_RELATIVE = 2

# Extension tags.
WTX_OBT = 0  # Out of bounds tracking
WTX_FKEYS = 1  # Function keys
WTX_TILT = 2  # Raw Cartesian tilt; 1.1
WTX_CSRMASK = 3  # select input by cursor type; 1.1
WTX_XBTNMASK = 4  # Extended button mask; 1.1
WTX_EXPKEYS = 5  # ExpressKeys; 1.3


def wtinfo(category, index, buffer):
    size = lib.WTInfoW(category, index, None)
    assert size <= ctypes.sizeof(buffer)
    lib.WTInfoW(category, index, ctypes.byref(buffer))
    return buffer


def wtinfo_string(category, index):
    size = lib.WTInfoW(category, index, None)
    buffer = ctypes.create_unicode_buffer(size)
    lib.WTInfoW(category, index, buffer)
    return buffer.value


def wtinfo_uint(category, index):
    buffer = UINT()
    lib.WTInfoW(category, index, ctypes.byref(buffer))
    return buffer.value


def wtinfo_word(category, index):
    buffer = WORD()
    lib.WTInfoW(category, index, ctypes.byref(buffer))
    return buffer.value


def wtinfo_dword(category, index):
    buffer = DWORD()
    lib.WTInfoW(category, index, ctypes.byref(buffer))
    return buffer.value


def wtinfo_wtpkt(category, index):
    buffer = WTPKT()
    lib.WTInfoW(category, index, ctypes.byref(buffer))
    return buffer.value


def wtinfo_bool(category, index):
    buffer = BOOL()
    lib.WTInfoW(category, index, ctypes.byref(buffer))
    return bool(buffer.value)


class Win32Wintab:
    '''High-level interface to tablet devices.

    Unlike other devices, tablets must be opened for a specific window,
    and cannot be opened exclusively.  The `open` method returns a
    `TabletCanvas` object, which supports the events provided by the tablet.

    Currently only one tablet device can be used, though it can be opened on
    multiple windows.  If more than one tablet is connected, the behaviour is
    undefined.
    '''

    def __init__(self, index):
        self._device = WTI_DEVICES + index
        self.name = wtinfo_string(self._device, DVC_NAME).strip()
        self.id = wtinfo_string(self._device, DVC_PNPID)
        self.hardware_type = wtinfo_uint(self._device, DVC_HARDWARE)

        self.hw_model_info = dict()
        self.hw_model_info['name'] = self.name
        self.hw_model_info['id'] = self.id
        self.hw_model_info['type'] = self.hardware_type
        self.hw_model_info['handle'] = self._device

        #phys_cursors = hardware & HWC_PHYSID_CURSORS

        n_cursors = wtinfo_uint(self._device, DVC_NCSRTYPES)
        first_cursor = wtinfo_uint(self._device, DVC_FIRSTCSR)

        self.hw_axis_info = dict()

        self._x_axis = wtinfo(self._device, DVC_X, AXIS())
        self._y_axis = wtinfo(self._device, DVC_Y, AXIS())
        self._z_axis = wtinfo(self._device, DVC_Z, AXIS())

        self.tip_pressure_axis = wtinfo(self._device, DVC_NPRESSURE,
                                        AXIS())
        # self.barrel_pressure_axis = wtinfo(self._device, DVC_TPRESSURE,
        #                            AXIS())

        self.orient_axis = wtinfo(self._device, DVC_ORIENTATION,
                                  (AXIS * 3)())

        # self.rotation_axis = wtinfo(self._device, DVC_ROTATION,
        #                          (AXIS*3)())

        self.hw_axis_info['x'] = struct2dict(self._x_axis)
        self.hw_axis_info['y'] = struct2dict(self._y_axis)
        self.hw_axis_info['z'] = struct2dict(self._z_axis)
        # seems that z axis min value is being reported as -1023,
        # but it is never < 0 in PACKET field, so setting z min to 0
        #print2err("z axis: ", self.hw_axis_info['z'])
        self.hw_axis_info['z']['min'] = 0

        self.hw_axis_info['pressure'] = struct2dict(self.tip_pressure_axis)
        #self.hw_axis_info['barrel_pressure_axis'] = struct2dict(self.barrel_pressure_axis)
        self.hw_axis_info['orient_azimuth'] = struct2dict(self.orient_axis[0])
        self.hw_axis_info['orient_altitude'] = struct2dict(self.orient_axis[1])
        self.hw_axis_info['orient_twist'] = struct2dict(self.orient_axis[2])
        #self.hw_axis_info['rotation_pitch_axis'] = struct2dict(self.rotation_axis[0])
        #self.hw_axis_info['rotation_roll_axis'] = struct2dict(self.rotation_axis[1])
        #self.hw_axis_info['rotation_yaw_axis'] = struct2dict(self.rotation_axis[2])

        self.cursors = []
        self._cursor_map = {}

        for i in range(n_cursors):
            cursor = Win32WintabCursor(self, i + first_cursor)
            if not cursor.bogus:
                self.cursors.append(cursor)
                self._cursor_map[i + first_cursor] = cursor

    def open(self, window, iohub_wt_device):
        """Open a tablet device for a window.

        :Parameters:
            `window` : `Window`
                The window on which the tablet will be used.

        :rtype: `Win32WintabCanvas`

        """
        Win32WintabCanvas.iohub_wt_device = iohub_wt_device
        pc = Win32WintabCanvas(self, window)
        return pc


class Win32WintabCanvas(EventDispatcher):
    """Event dispatcher for tablets.

    Use `Tablet.open` to obtain this object for a particular tablet device and
    window.  Events may be generated even if the tablet stylus is outside of
    the window; this is operating-system dependent.

    The events each provide the `TabletCursor` that was used to generate the
    event; for example, to distinguish between a stylus and an eraser.  Only
    one cursor can be used at a time, otherwise the results are undefined.

    :Ivariables:
        `window` : Window
            The window on which this tablet was opened.

    """
    iohub_wt_device = None

    def __init__(self, device, window, msg_base=WT_DEFBASE):
        super(Win32WintabCanvas, self).__init__()
        self.window = window

        self.device = device
        self.msg_base = msg_base

        self._iohub_events = []

        # TODO: Let user determine if raw or pixel coords should be used for
        #       x,y pos fields. i.e. WTI_DEFCONTEXT == raw,
        #       WTI_DEFSYSCTX == pix
        self.context_info = context_info = LOGCONTEXT()
        #wtinfo(WTI_DEFSYSCTX, 0, context_info)
        wtinfo(WTI_DEFCONTEXT, 0, context_info)
        context_info.lcMsgBase = msg_base
        context_info.lcOptions |= CXO_MESSAGES

        # If you change this, change definition of PACKET also.
        context_info.lcPktData = DEFAULT_PACKET_DATA_FIELDS
        context_info.lcPktMode = 0   # All absolute

        # set output scaling, right now, out == in
        # TODO: determine what coord space we want to output in.
        context_info.lcOutOrgX = context_info.lcOutOrgY = 0

        # Set the entire tablet as active
        context_info.lcInOrgX = 0
        context_info.lcInOrgY = 0
        context_info.lcInOrgZ = 0
        context_info.lcInExtX = self.device._x_axis.axMax
        context_info.lcInExtY = self.device._y_axis.axMax
        context_info.lcInExtZ = self.device._z_axis.axMax

        context_info.lcOutOrgX = context_info.lcInOrgX
        context_info.lcOutOrgY = context_info.lcInOrgY
        context_info.lcOutOrgZ = context_info.lcInOrgZ
        # INT(scale[0] * context_info.lcInExtX);
        context_info.lcOutExtX = context_info.lcInExtX
        # INT(scale[1] * context_info.lcInExtY);
        context_info.lcOutExtY = context_info.lcInExtY
        # INT(scale[1] * context_info.lcInExtY);
        context_info.lcOutExtZ = context_info.lcInExtZ

        self._context = lib.WTOpenW(window._hwnd,
                                    ctypes.byref(context_info),
                                    self.iohub_wt_device.getConfiguration().
                                    get('auto_report_events', False))
        if not self._context:
            print2err("Couldn't open tablet context.")
        window._event_handlers[msg_base + WT_PACKET] = self._event_wt_packet
        window._event_handlers[msg_base + WT_PROXIMITY] = self._event_wt_proximity

        self._current_cursor = None
        #self._pressure_scale = device.tip_pressure_axis.get_scale()
        #self._pressure_bias = device.tip_pressure_axis.get_bias()

    def getContextInfo(self):
        return struct2dict(self.context_info)

    def enable(self, enable=True):
        lib.WTEnable(self._context, enable)

    def close(self):
        lib.WTClose(self._context)
        self._context = None
        del self.window._event_handlers[self.msg_base + WT_PACKET]
        del self.window._event_handlers[self.msg_base + WT_PROXIMITY]

    def _set_current_cursor(self, cursor_type):
        self._current_cursor = self.device._cursor_map.get(cursor_type, None)

        if self._current_cursor:
            self.addHubEvent(None, EventConstants.WINTAB_ENTER_REGION)

    def addHubEvent(
            self,
            packet,
            evt_type=EventConstants.WINTAB_SAMPLE):
        # TODO: Display (screen) index
        #display_id = 0
        #window_id = self.window._hwnd

        #ccursor = self._current_cursor
        #if ccursor is None:
        #    ccursor = 0
        #else:
        #    ccursor = ccursor._cursor

        if packet is None or evt_type != EventConstants.WINTAB_SAMPLE:
            cevt = [evt_type,
                    0,
                    0,
                    0,
                    0,
                    0,  # ('x',N.int32),
                    0,  # ('y',N.int32),
                    0,  # ('z',N.int32),
                    0,  # ('pressure',N.uint32),
                    0,  # ('orient_azimuth',N.int32),
                    0,  # ('orient_altitude;',N.int32),
                    0,  # ('orient_twist',N.int32),
                    0  # ('status', N.uint8)
                    ]
        else:
            cevt = [evt_type,
                    packet.pkTime / 1000.0,
                    packet.pkStatus,
                    packet.pkSerialNumber,
                    packet.pkButtons,
                    packet.pkX,  # ('x',N.int32),
                    packet.pkY,  # ('y',N.int32),
                    packet.pkZ,  # ('z',N.int32),
                    #(packet.pkNormalPressure + self._pressure_bias) * \
                    #            self._pressure_scale,
                    packet.pkNormalPressure,  # ('pressure',N.uint32),
                    # ('orient_azimuth',N.int32),
                    packet.pkOrientation.orAzimuth,
                    # ('orient_altitude;',N.int32),
                    packet.pkOrientation.orAltitude,
                    packet.pkOrientation.orTwist,  # ('orient_twist',N.int32),
                    0  # ('status', N.uint8)
                    ]
        self._iohub_events.append(cevt)

    @pyglet.window.win32.Win32EventHandler(0)
    def _event_wt_packet(self, msg, wParam, lParam):
        if lParam != self._context:
            return

        packet = PACKET()
        if lib.WTPacket(self._context, wParam, ctypes.byref(packet)) == 0:
            return

        if not packet.pkChanged:
            return

        if self._current_cursor is None:
            self._set_current_cursor(packet.pkCursor)

        self.addHubEvent(packet)

    @pyglet.window.win32.Win32EventHandler(0)
    def _event_wt_proximity(self, msg, wParam, lParam):
        if wParam != self._context:
            return

        if lParam & 0xffff:
            # Proximity Enter Event
            # If going in, proximity event will be generated by next event, which
            # can actually grab a cursor id, so evt is generated when
            # _set_current_cursor() is called with a non None cursor.
            pass
        else:
            # Proximity Leave Event
            self.addHubEvent(None, EventConstants.WINTAB_LEAVE_REGION)
            self._current_cursor = None


class Win32WintabCursor:

    def __init__(self, device, index):
        self.device = device
        self._cursor = WTI_CURSORS + index

        self.name = wtinfo_string(self._cursor, CSR_NAME).strip()
        self.active = wtinfo_bool(self._cursor, CSR_ACTIVE)
        pktdata = wtinfo_wtpkt(self._cursor, CSR_PKTDATA)

        # A whole bunch of cursors are reported by the driver, but most of
        # them are hogwash.  Make sure a cursor has at least X and Y data
        # before adding it to the device.
        self.bogus = not (pktdata & PK_X and pktdata & PK_Y)
        if self.bogus:
            return

        self.id = (wtinfo_dword(self._cursor, CSR_TYPE) << 32) | \
            wtinfo_dword(self._cursor, CSR_PHYSID)

    def __repr__(self):
        return 'WintabCursor(%r)' % self.name


def get_spec_version():
    spec_version = wtinfo_word(WTI_INTERFACE, IFC_SPECVERSION)
    return spec_version


def get_interface_name():
    interface_name = wtinfo_string(WTI_INTERFACE, IFC_WINTABID)
    return interface_name


def get_implementation_version():
    impl_version = wtinfo_word(WTI_INTERFACE, IFC_IMPLVERSION)
    return impl_version


def get_tablets():
    # Require spec version 1.1 or greater
    if get_spec_version() < 0x101:
        return []

    n_devices = wtinfo_uint(WTI_INTERFACE, IFC_NDEVICES)
    devices = [Win32Wintab(i) for i in range(n_devices)]
    return devices
